/*
 * GIMP Speedline Plug-in -- This is a plug-in for The GIMP 2
 * Copyright (C) 2013-2015 OHTSUKA, Yoshio <ohtsuka.yoshio@gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */
#include "speedline.h"
#include "speedline_preview.h"
#include "speedline_radiation.h"
#include "speedline_parallel.h"

typedef enum _PlguinError {
	ERROR_NOTHING, 
	ERROR_CANCEL,
	ERROR_FAILED_CREATE_LAYER_ERROR,
	ERROR_UNKNOWN_IMAGE_TYPE_ERROR,
	ERROR_UNKNOWN_ERROR
} PluginError;

/*static void init    (void);*/
/*static void quit    (void);*/
static void query	(void);
static void run		(const gchar		*name,
					gint				nparams,
					const GimpParam		*param,
					gint				*nreturn_vals,
					GimpParam			**return_vals);

GimpPlugInInfo PLUG_IN_INFO = {
	NULL, /* init */
	NULL, /* quit */
	query,
	run
};

#define SPEEDLINE_RESPONSE_CLOSE   0
#define SPEEDLINE_RESPONSE_CREATE  1
#define SPEEDLINE_RESPONSE_MODIFY  2
#define SPEEDLINE_RESPONSE_RESET   3
#define SPEEDLINE_RESPONSE_TOPMOST 4

DialogWidgets widgets;

// template vals, used when modify mode
GKeyFile *tempvals;

// key name of parameter
const static char PNAME_GROUP[]                               = PLUGIN_NAME;
const static char PNAME_VERSION[]                             = "version";
const static char PNAME_LINE_COLOR[]                          = "line_color";
const static char PNAME_LINE_ANTIALIAS[]                      = "line_antialias";
const static char PNAME_LINE_PITCH[]                          = "line_pitch";
const static char PNAME_LINE_PITCH_RAND[]                     = "line_pitch_rand";
const static char PNAME_LINE_PITCH_RAND_SEED[]                = "line_pitch_rand_seed";
const static char PNAME_LINE_PITCH_TYPE[]                     = "line_pitch_type";
const static char PNAME_LINE_BLOCK_ON_PITCH[]                 = "line_block_on_pitch";
const static char PNAME_LINE_BLOCK_ON_PITCH_RAND[]            = "line_block_on_pitch_rand";
const static char PNAME_LINE_BLOCK_ON_PITCH_RAND_SEED[]       = "line_blick_on_pitch_rand_seed";
const static char PNAME_LINE_BLOCK_OFF_PITCH[]                = "line_block_off_pitch";
const static char PNAME_LINE_BLOCK_OFF_PITCH_RAND[]           = "line_block_off_pitch_rand";
const static char PNAME_LINE_BLOCK_OFF_PITCH_RAND_SEED[]      = "line_blick_off_pitch_rand_seed";
const static char PNAME_LINE_WAVE_NPOINT[]                    = "line_wave_npoint";
const static char PNAME_LINE_WAVE_DEPTH[]                     = "line_wave_depth";
const static char PNAME_LINE_WAVE_DEPTH_RAND[]                = "line_wave_depth_rand";
const static char PNAME_LINE_WAVE_DEPTH_RAND_SEED[]           = "line_wave_depth_rand_seed";
const static char PNAME_LINE_WAVE_TYPE[]                      = "line_wave_type";
const static char PNAME_LINE_WIDTH[]                          = "line_width";
const static char PNAME_LINE_WIDTH_RAND[]                     = "line_width_rand";
const static char PNAME_LINE_WIDTH_RAND_SEED[]                = "line_width_rand_seed";
const static char PNAME_LINE_IN_MARGIN[]                      = "line_in_margin";
const static char PNAME_LINE_IN_MARGIN_RAND[]                 = "line_in_margin_rand";
const static char PNAME_LINE_IN_MARGIN_RAND_SEED[]            = "line_in_margin_rand_seed";
const static char PNAME_LINE_IN_LENGTH[]                      = "line_in_length";
const static char PNAME_LINE_IN_LENGTH_RAND[]                 = "line_in_length_rand";
const static char PNAME_LINE_IN_LENGTH_RAND_SEED[]            = "line_in_length_rand_seed";
const static char PNAME_LINE_OUT_MARGIN[]                     = "line_out_margin";
const static char PNAME_LINE_OUT_MARGIN_RAND[]                = "line_out_margin_rand";
const static char PNAME_LINE_OUT_MARGIN_RAND_SEED[]           = "line_out_margin_rand_seed";
const static char PNAME_LINE_OUT_LENGTH[]                     = "line_out_length";
const static char PNAME_LINE_OUT_LENGTH_RAND[]                = "line_out_length_rand";
const static char PNAME_LINE_OUT_LENGTH_RAND_SEED[]           = "line_out_length_rand_seed";
const static char PNAME_AREA_TYPE[]                           = "area_type";
const static char PNAME_AREA_UNIT[]                           = "area_unit";
const static char PNAME_RADIATION_X[]                         = "radiation_x";
const static char PNAME_RADIATION_Y[]                         = "radiation_y";
const static char PNAME_RADIATION_IN_RADIUS[]                 = "radiation_in_radius";
const static char PNAME_RADIATION_OUT_RADIUS[]                = "radiation_out_radius";
const static char PNAME_RADIATION_OBLATENESS[]                = "radiation_oblateness";
const static char PNAME_RADIATION_IN_FREEHAND_ACTIVE[]        = "radiation_in_freehand_active";
const static char PNAME_RADIATION_IN_FREEHAND_POINTS[]        = "radiation_in_freehand_points";
const static char PNAME_RADIATION_IN_FREEHAND_POINTS_COUNT[]  = "radiation_in_freehand_points_count";
const static char PNAME_RADIATION_OUT_FREEHAND_ACTIVE[]       = "radiation_out_freehand_active";
const static char PNAME_RADIATION_OUT_FREEHAND_POINTS[]       = "radiation_out_freehand_points";
const static char PNAME_RADIATION_OUT_FREEHAND_POINTS_COUNT[] = "radiation_out_freehand_points_count";
const static char PNAME_PARALLEL_IN_X[]                       = "parallel_in_x";
const static char PNAME_PARALLEL_IN_Y[]                       = "parallel_in_Y";
const static char PNAME_PARALLEL_IN_RANGE[]                   = "parallel_in_range";
const static char PNAME_PARALLEL_OUT_X[]                      = "parallel_out_x";
const static char PNAME_PARALLEL_OUT_Y[]                      = "parallel_out_Y";
const static char PNAME_PARALLEL_OUT_RANGE[]                  = "parallel_out_range";
const static char PNAME_PARALLEL_IN_FREEHAND_ACTIVE[]         = "parallel_in_freehand_active";
const static char PNAME_PARALLEL_IN_FREEHAND_POINTS[]         = "parallel_in_freehand_points";
const static char PNAME_PARALLEL_IN_FREEHAND_POINTS_COUNT[]   = "parallel_in_freehand_points_count";
const static char PNAME_PARALLEL_OUT_FREEHAND_ACTIVE[]        = "parallel_out_freehand_active";
const static char PNAME_PARALLEL_OUT_FREEHAND_POINTS[]        = "parallel_out_freehand_points";
const static char PNAME_PARALLEL_OUT_FREEHAND_POINTS_COUNT[]  = "parallel_out_freehand_points_count";

static GtkWidget* build_gtk_spin_button_new (double min, double max, double step, guint digits, double value);
static GtkWidget* build_line_antialias_combo_box_new (AntialiasType default_value);
static GtkWidget* build_line_pitch_type_combo_box_new (LinePitchType default_value);
static GtkWidget* build_line_wave_type_combo_box_new (LineWaveType default_value);
static GtkWidget* build_unit_combo_box_new (gboolean has_percent, GimpUnit default_value);

static int run_dialog (PluginVals *vals);
static int do_speedline (PluginVals *vals, gint32 *out_layer_ID);

static char* create_layer_name (char *text);
static void reset_default_vals (PluginVals *vals);

static void preset_load (gchar *path);
static void preset_save (gchar *path, PluginVals *vals);

static void          parasite_to_widgets      (GimpParasite *parasite);
static GimpParasite* pluginvals_to_parasite   (PluginVals *vals);

static GKeyFile*   pluginvals_to_keyfile    (PluginVals *vals);
static void        pluginvals_to_widgets    (PluginVals *vals);
static PluginVals* keyfile_to_pluginvals    (GKeyFile *kfile);
static void        keyfile_to_widgets       (GKeyFile *kfile);
//Globals
static gchar *preset_dir = NULL;


/*
 Data Flow of Parameter
                                                                                       
                            +--------------+      +------------+      +-------------+  
                            |  GtkWidgets  |      |  GKeyFile  +----->|  Parasite   |  
                            +------+-------+      +---+------+-+<-----+-------------+  
                                   |  A             A |    A |                         
                                   V  |             | |    | |                         
                            +---------+----+        | |    | |        +-------------+  
  INTERACTIVE     --------->|  PluginVals  +--------+ |    | +------->|    Preset   | 
  NON-INTERACTIVE           +------+-------+<---------+    +----------+-------------+  
                                   |                                                   
                                   V                                                   
                            +--------------+                                           
                            |   Renderer   |                                           
                            +--------------+                                           
                                                                                       
 */

// Gimp Plugin Interface
static void
query (void) {
	static GimpParamDef args[] = {
	{ GIMP_PDB_INT32,   "run-mode",     "Interactive=0, non-interactive=1" },
	{ GIMP_PDB_IMAGE,   "image",        "Input image" },
	{ GIMP_PDB_DRAWABLE,"drawable",     "Input drawable" },
	};
	
	static GimpParamDef return_vals[] = {
	{ GIMP_PDB_LAYER,	"layer",	"The speedline layer" }
	};
	
	gimp_install_procedure (
		PLUGIN_ID,
		_(PLUGIN_DESCRIPTION),
		_(PLUGIN_DESCRIPTION),
		PLUGIN_AUTHOR,
		PLUGIN_AUTHOR,
		PLUGIN_DATE,
		_("_Speedline..."),
		"RGB*, GRAY*",
		GIMP_PLUGIN,
		G_N_ELEMENTS (args), G_N_ELEMENTS (return_vals),
		args, return_vals);

	gimp_plugin_menu_register (PLUGIN_ID, "<Image>/Filters/ReddogPlug/");
}


static void
run (const gchar	*name,
	gint			nparams,
	const GimpParam	*param,
	gint			*nreturn_vals,
	GimpParam		**return_vals)
{/*{{{*/
	static GimpParam	values[2];
    //GimpDrawable		*drawable;
	gint32				image_ID;
	gint32				layer_ID;
	GimpRunMode			run_mode;
	GimpPDBStatusType	status = GIMP_PDB_SUCCESS;
	
	PluginVals vals;
	
	INIT_I18N();

	*nreturn_vals = 2;
	*return_vals  = values;
	
	run_mode = param[0].data.d_int32;
	image_ID = param[1].data.d_image;
	layer_ID = param[2].data.d_drawable;
	
	//return value
	values[0].type          = GIMP_PDB_STATUS;
	values[0].data.d_status = status;
	values[1].type          = GIMP_PDB_LAYER;
	values[1].data.d_int32  = -1;

	//init parameter
	vals.image_ID           = image_ID;
	vals.layer_ID           = layer_ID;

	//save dialog parameter
	vals.gui.x           = 0;
	vals.gui.y           = 0;
	vals.gui.width       = -1;
	vals.gui.height      = -1;
	vals.gui.topmost     = FALSE;
	vals.gui.page_number = 0;
	
	// init run mode
	vals.mode = RUNMODE_ADD;

	// default parameters value
	reset_default_vals (&vals);

	switch (run_mode) {
		//interactive mode
		case GIMP_RUN_INTERACTIVE:
		case GIMP_RUN_WITH_LAST_VALS:
		{
			// Restore Parameters
			gimp_get_data (PLUGIN_ID, &vals);
			vals.image_ID = image_ID;
			vals.layer_ID = layer_ID;
			vals.mode     = RUNMODE_ADD;

			//run dialog and enter message loops
			if (ERROR_NOTHING != run_dialog (&vals)) {
				WNGMSG ("Error: Failed to get params from GUI interface");
				status = GIMP_PDB_CALLING_ERROR;
				goto error;
			}

			//Store Parameters
			gimp_set_data (PLUGIN_ID, &vals, sizeof (PluginVals));

			values[0].type			= GIMP_PDB_SUCCESS;
			values[1].data.d_int32	= -1;
		}
		break;

		// non-interactive mode
		case GIMP_RUN_NONINTERACTIVE:
            WNGMSG (_("error: This plug-in does not support non-interactive mode."));
            status = GIMP_PDB_CALLING_ERROR;
            goto error;
			break;
	}

    //gimp_drawable_detach (drawable);
	return;

error:
    //gimp_drawable_detach (drawable);
	values[0].data.d_status = status;
	values[1].type          = GIMP_PDB_LAYER;
	values[1].data.d_int32	= -1;
	return;
}/*}}}*/

static gchar *
get_preset_dir ()
{/*{{{*/
	gchar *ret        = NULL;
	gchar *dir        = NULL;
	gchar *parent_dir = NULL;
	gchar *gimprc     = gimp_personal_rc_file ("gimprc");
	if (gimprc == NULL) goto done;

	parent_dir = g_path_get_dirname (gimprc);
	dir = g_build_filename (parent_dir, PLUGIN_NAME, NULL);
	
	if (g_file_test (dir, G_FILE_TEST_EXISTS) == FALSE) {
		if (g_mkdir (dir, 0755) != 0) {
			if (dir) g_free (dir);
			goto done;
		}
	} else if (g_file_test (dir, G_FILE_TEST_IS_DIR) == FALSE) {
		if (dir) g_free (dir);
		goto done;
	}
	ret = dir;
done:
	if (parent_dir) g_free (parent_dir);
	if (gimprc) g_free (gimprc);
	return ret;
}/*}}}*/

static void
response_callback (GtkWidget	*dialog,
				   gint			response_id,
				   PluginVals	*vals)
{/*{{{*/
	GtkWidget *add_button;
	GtkWidget *close_button;
	GtkWidget *modify_button;
	GtkWidget *reset_button;

	add_button = gtk_dialog_get_widget_for_response (GTK_DIALOG (dialog), SPEEDLINE_RESPONSE_CREATE);
	close_button = gtk_dialog_get_widget_for_response (GTK_DIALOG (dialog), SPEEDLINE_RESPONSE_CLOSE);
	modify_button = gtk_dialog_get_widget_for_response (GTK_DIALOG (dialog), SPEEDLINE_RESPONSE_MODIFY);
	reset_button = gtk_dialog_get_widget_for_response (GTK_DIALOG (dialog), SPEEDLINE_RESPONSE_RESET);


	switch (response_id) {
		case SPEEDLINE_RESPONSE_CLOSE:
		{
			switch (vals->mode) {
				case  RUNMODE_MODIFY:
				{
					// restore parameter
					if (tempvals) {
						keyfile_to_widgets (tempvals);
						if (tempvals) {
							g_key_file_free (tempvals);
							tempvals = NULL;
						}
					}

					vals->mode = RUNMODE_ADD;
					gtk_button_set_label (GTK_BUTTON (add_button), _("Create"));
					gtk_button_set_label (GTK_BUTTON (close_button), _("Close"));
					gtk_widget_set_visible (modify_button, TRUE);
					break;
				}
				case RUNMODE_ADD:
				{
					gtk_window_get_size (GTK_WINDOW (dialog), &vals->gui.width, &vals->gui.height);
					gtk_window_get_position (GTK_WINDOW (dialog), &vals->gui.x, &vals->gui.y);
					gtk_widget_destroy (dialog);
					break;
				}
				default:
					break;
			}
			break;
		}

		case SPEEDLINE_RESPONSE_CREATE:
		{
			switch (vals->mode) {
				case RUNMODE_MODIFY:
				{
					// do_speedline as modify mode
					gint32 out_layer_ID;
					do_speedline (vals, &out_layer_ID);
					// restore parameter
					if (tempvals) {
						keyfile_to_widgets (tempvals);
						g_key_file_free (tempvals);
						tempvals = NULL;
					}

					vals->mode = RUNMODE_ADD;
					gtk_button_set_label (GTK_BUTTON (add_button), _("Create"));
					gtk_button_set_label (GTK_BUTTON (close_button), _("Close"));
					gtk_widget_set_visible (modify_button, TRUE);
					break;
				}
				case RUNMODE_ADD: 
				{
					gint32 out_layer_ID;
					do_speedline (vals, &out_layer_ID);

					response_callback (dialog, SPEEDLINE_RESPONSE_CLOSE, vals);
					break;
				}
				default:
					break;
			}
			break;
		}
		

		case SPEEDLINE_RESPONSE_MODIFY:
		{
			switch (vals->mode) {
				case RUNMODE_ADD:
				{
					gint32 active_layer;
					GimpParasite *parasite;

					// check if this layer has parasite
					active_layer = gimp_image_get_active_layer (vals->image_ID);
					parasite = gimp_item_get_parasite (active_layer, PLUGIN_ID);

					if (parasite != NULL) {
						if (tempvals) {
							g_key_file_free (tempvals);
							tempvals = NULL;
						}
						tempvals = pluginvals_to_keyfile (vals);
						
						//load parameters from parasite which attached on layer
						parasite_to_widgets (parasite);
						gimp_parasite_free (parasite);
						
						gtk_button_set_label (GTK_BUTTON (add_button), _("Apply"));
						gtk_button_set_label (GTK_BUTTON (close_button), _("Cancel"));
						gtk_widget_set_visible (modify_button, FALSE);
						vals->layer_ID = active_layer;
						vals->mode = RUNMODE_MODIFY;
					} else {
						msgbox (GTK_MESSAGE_ERROR, GTK_BUTTONS_CLOSE, NULL,
							_("Active layer is not a Speedline Layer!"));
					}
					break;
				}
				case RUNMODE_MODIFY:
				{
					// invisible when modify mode
					break;
				}
				default:
					break;
			}
		}

		case SPEEDLINE_RESPONSE_RESET:
			reset_default_vals (vals);
			pluginvals_to_widgets (vals);
			gtk_widget_queue_draw (widgets.preview);
			break;

		case SPEEDLINE_RESPONSE_TOPMOST:
		{
			static gboolean topmost = FALSE;
			topmost = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widgets.topmost));
			gtk_window_set_keep_above (GTK_WINDOW (dialog), topmost);
			vals->gui.topmost = topmost;
			break;
		}
	}
}/*}}}*/

static void
preview_controller_button_clicked_callback (GtkWidget	*widget,
											PluginVals	*vals)
{/*}}}*/
	if (widget == widgets.preview_plus_button) {
		preview_zoom_in ();
	}
	if (widget == widgets.preview_minus_button) {
		preview_zoom_out ();
	}
	//gtk_widget_queue_draw (widgets.preview);
}/*}}}*/

static void basenotebook_callback (GtkNotebook		*notebook,
								   GtkNotebookPage	*page,
								   guint			page_num,
								   PluginVals		*vals)
{/*{{{*/
	vals->gui.page_number = page_num;
	DBGMSG ("speedline gui: selected notebook page %d\n", vals->gui.page_number);
}/*}}}*/

static void
area_type_radio_click_callback (GtkWidget	*widget,
								PluginVals	*vals)
{/*}}}*/
	if (widget == widgets.area_type_radiation) {
		vals->area_type = AREA_TYPE_RADIATION;
	}
	if (widget == widgets.area_type_parallel) {
		vals->area_type = AREA_TYPE_PARALLEL;
	}
	if (widgets.area_type_notebook != NULL) {
		gtk_notebook_set_current_page (GTK_NOTEBOOK (widgets.area_type_notebook), vals->area_type);
	}
	gtk_widget_queue_draw (widgets.preview);
}/*}}}*/

static void
area_type_unit_changed_callback (GtkComboBox *widget,
								 PluginVals  *vals)
{/*}}}*/
	static GimpUnit beforeUnit = GIMP_UNIT_PIXEL;
#define _CONVERT(VAL, RES) \
{ \
	gdouble pixels;\
	if (beforeUnit == GIMP_UNIT_PIXEL) {\
		pixels = (VAL);\
	} else {\
		pixels = gimp_units_to_pixels ((VAL), beforeUnit, (RES));\
	}\
	(VAL) = gimp_pixels_to_units (pixels, vals->area_unit, (RES));\
}

	vals->area_unit = gtk_combo_box_get_active (widget);

	//convert parameters unit.
	if (vals->area_unit != beforeUnit) {
		gdouble resx, resy;
		gimp_image_get_resolution (vals->image_ID, &resx, &resy);

		_CONVERT (vals->line_pitch, resx);
		_CONVERT (vals->line_block_on_pitch, resx);
		_CONVERT (vals->line_block_off_pitch, resx);
		_CONVERT (vals->line_wave_depth, resx);
		_CONVERT (vals->line_width, resx);
		_CONVERT (vals->line_in_margin, resx);
		_CONVERT (vals->line_in_length, resx);
		_CONVERT (vals->line_out_margin, resx);
		_CONVERT (vals->line_out_length, resx);

		_CONVERT (vals->radiation_x, resx);
		_CONVERT (vals->radiation_y, resy);
		_CONVERT (vals->radiation_in_radius, resx);
		_CONVERT (vals->radiation_out_radius, resx);

		_CONVERT (vals->parallel_in_x, resx);
		_CONVERT (vals->parallel_in_y, resy);
		_CONVERT (vals->parallel_in_range, resx);
		_CONVERT (vals->parallel_out_x, resx);
		_CONVERT (vals->parallel_out_y, resy);
		_CONVERT (vals->parallel_out_range, resx);

		pluginvals_to_widgets (vals);

		beforeUnit = vals->area_unit;
	}
#undef _CONVERT
}/*}}}*/

static void
radiation_x_value_changed_callback (GtkSpinButton	*widget,
									PluginVals		*vals)
{/*}}}*/
	vals->radiation_x = gtk_spin_button_get_value (widget);
	gtk_widget_queue_draw (widgets.preview);
}/*}}}*/

static void
radiation_y_value_changed_callback (GtkSpinButton	*widget,
									PluginVals		*vals)
{/*}}}*/
	vals->radiation_y = gtk_spin_button_get_value (widget);
	gtk_widget_queue_draw (widgets.preview);
}/*}}}*/

static void
radiation_in_radius_value_changed_callback (GtkSpinButton	*widget,
											PluginVals		*vals)
{/*}}}*/
	vals->radiation_in_radius = gtk_spin_button_get_value (widget);
	gtk_widget_queue_draw (widgets.preview);
}/*}}}*/

static void
radiation_out_radius_value_changed_callback (GtkSpinButton	*widget,
											 PluginVals		*vals)
{/*}}}*/
	vals->radiation_out_radius = gtk_spin_button_get_value (widget);
	gtk_widget_queue_draw (widgets.preview);
}/*}}}*/

static void
radiation_oblateness_value_changed_callback (GtkSpinButton	*widget,
											 PluginVals		*vals)
{/*}}}*/
	vals->radiation_oblateness = gtk_spin_button_get_value (widget);
	gtk_widget_queue_draw (widgets.preview);
}/*}}}*/

static void
parallel_in_x_value_changed_callback (GtkSpinButton	*widget,
									  PluginVals	*vals)
{/*}}}*/
	vals->parallel_in_x = gtk_spin_button_get_value (widget);
	gtk_widget_queue_draw (widgets.preview);
}/*}}}*/

static void
parallel_in_y_value_changed_callback (GtkSpinButton	*widget,
									  PluginVals	*vals)
{/*}}}*/
	vals->parallel_in_y = gtk_spin_button_get_value (widget);
	gtk_widget_queue_draw (widgets.preview);
}/*}}}*/

static void
parallel_in_range_value_changed_callback (GtkSpinButton	*widget,
										  PluginVals	*vals)
{/*}}}*/
	vals->parallel_in_range = gtk_spin_button_get_value (widget);
	gtk_widget_queue_draw (widgets.preview);
}/*}}}*/

static void
parallel_out_x_value_changed_callback (GtkSpinButton	*widget,
									   PluginVals		*vals)
{/*}}}*/
	vals->parallel_out_x = gtk_spin_button_get_value (widget);
	gtk_widget_queue_draw (widgets.preview);
}/*}}}*/

static void
parallel_out_y_value_changed_callback (GtkSpinButton	*widget,
									   PluginVals		*vals)
{/*}}}*/
	vals->parallel_out_y = gtk_spin_button_get_value (widget);
	gtk_widget_queue_draw (widgets.preview);
}/*}}}*/

static void
parallel_out_range_value_changed_callback (GtkSpinButton	*widget,
										   PluginVals		*vals)
{/*}}}*/
	vals->parallel_out_range = gtk_spin_button_get_value (widget);
	gtk_widget_queue_draw (widgets.preview);
}/*}}}*/

static void
line_color_color_changed_callback (GtkWidget	*widget,
								   PluginVals	*vals)
{/*}}}*/
	gimp_color_button_get_color (GIMP_COLOR_BUTTON (widget), &vals->line_color);
	gtk_widget_queue_draw (widgets.preview);
}/*}}}*/

static void
line_antialias_changed_callback (GtkComboBox	*widget,
								 PluginVals		*vals)
{/*}}}*/
	AntialiasType type;
	gimp_int_combo_box_get_active (GIMP_INT_COMBO_BOX (widget), (int*)&type);
	vals->line_antialias = type;
	gtk_widget_queue_draw (widgets.preview);
}/*}}}*/

static void
line_pitch_value_changed_callback (GtkSpinButton	*widget,
								   PluginVals		*vals)
{/*}}}*/
	vals->line_pitch = gtk_spin_button_get_value (widget);
	gtk_widget_queue_draw (widgets.preview);
}/*}}}*/

static void
line_pitch_rand_value_changed_callback (GtkSpinButton	*widget,
										PluginVals		*vals)
{/*}}}*/
	vals->line_pitch_rand = gtk_spin_button_get_value (widget);
	gtk_widget_queue_draw (widgets.preview);
}/*}}}*/

static void
line_pitch_rand_seed_value_changed_callback (GtkSpinButton	*widget,
											 PluginVals		*vals)
{/*}}}*/
	vals->line_pitch_rand_seed = (gint)gtk_spin_button_get_value (widget);
	gtk_widget_queue_draw (widgets.preview);
}/*}}}*/

static void
line_pitch_type_changed_callback (GimpIntComboBox	*widget,
								  PluginVals		*vals)
{/*}}}*/
	LinePitchType type;
	gimp_int_combo_box_get_active (GIMP_INT_COMBO_BOX (widget), (int*)&type);
	vals->line_pitch_type = type;
	gtk_widget_queue_draw (widgets.preview);
}/*}}}*/

static void
line_block_on_pitch_value_changed_callback (GtkSpinButton	*widget,
											PluginVals		*vals)
{/*}}}*/
	vals->line_block_on_pitch = gtk_spin_button_get_value (widget);
	gtk_widget_queue_draw (widgets.preview);
}/*}}}*/

static void
line_block_on_pitch_rand_value_changed_callback (GtkSpinButton	*widget,
												 PluginVals		*vals)
{/*}}}*/
	vals->line_block_on_pitch_rand = gtk_spin_button_get_value (widget);
	gtk_widget_queue_draw (widgets.preview);
}/*}}}*/

static void
line_block_on_pitch_rand_seed_value_changed_callback (GtkSpinButton	*widget,
													  PluginVals	*vals)
{/*}}}*/
	vals->line_block_on_pitch_rand_seed = (gint)gtk_spin_button_get_value (widget);
	gtk_widget_queue_draw (widgets.preview);
}/*}}}*/

static void
line_block_off_pitch_value_changed_callback (GtkSpinButton	*widget,
											 PluginVals		*vals)
{/*}}}*/
	vals->line_block_off_pitch = gtk_spin_button_get_value (widget);
	gtk_widget_queue_draw (widgets.preview);
}/*}}}*/

static void
line_block_off_pitch_rand_value_changed_callback (GtkSpinButton	*widget,
												  PluginVals	*vals)
{/*}}}*/
	vals->line_block_off_pitch_rand = gtk_spin_button_get_value (widget);
	gtk_widget_queue_draw (widgets.preview);
}/*}}}*/

static void
line_block_off_pitch_rand_seed_value_changed_callback (GtkSpinButton	*widget,
													   PluginVals		*vals)
{/*}}}*/
	vals->line_block_off_pitch_rand_seed = (gint)gtk_spin_button_get_value (widget);
	gtk_widget_queue_draw (widgets.preview);
}/*}}}*/


static void
line_wave_npoint_value_changed_callback (GtkSpinButton	*widget,
										 PluginVals		*vals)
{/*}}}*/
	int n =  (int)gtk_spin_button_get_value (widget);
	if ((n % 4) != 0) {
		gtk_spin_button_set_value (widget, n - (n % 4));
		return;
	}
	vals->line_wave_npoint = n;
	gtk_widget_queue_draw (widgets.preview);
}/*}}}*/

static void
line_wave_depth_value_changed_callback (GtkSpinButton	*widget,
										PluginVals		*vals)
{/*}}}*/
	vals->line_wave_depth = gtk_spin_button_get_value (widget);
	gtk_widget_queue_draw (widgets.preview);
}/*}}}*/

static void
line_wave_depth_rand_value_changed_callback (GtkSpinButton	*widget,
											 PluginVals		*vals)
{/*}}}*/
	vals->line_wave_depth_rand = gtk_spin_button_get_value (widget);
	gtk_widget_queue_draw (widgets.preview);
}/*}}}*/

static void
line_wave_depth_rand_seed_value_changed_callback (GtkSpinButton	*widget,
												  PluginVals	*vals)
{/*}}}*/
	vals->line_wave_depth_rand_seed = (gint)gtk_spin_button_get_value (widget);
	gtk_widget_queue_draw (widgets.preview);
}/*}}}*/

static void
line_wave_type_changed_callback (GimpIntComboBox	*widget,
								 PluginVals			*vals)
{/*}}}*/
	LineWaveType type;
	gimp_int_combo_box_get_active (GIMP_INT_COMBO_BOX (widget), (int*)&type);
	vals->line_wave_type = type;
	gtk_widget_queue_draw (widgets.preview);
}/*}}}*/

static void
line_width_value_changed_callback (GtkSpinButton	*widget,
								   PluginVals		*vals)
{/*}}}*/
	vals->line_width = gtk_spin_button_get_value (widget);
	gtk_widget_queue_draw (widgets.preview);
}/*}}}*/

static void
line_width_rand_value_changed_callback (GtkSpinButton	*widget,
										PluginVals		*vals)
{/*}}}*/
	vals->line_width_rand = gtk_spin_button_get_value (widget);
	gtk_widget_queue_draw (widgets.preview);
}/*}}}*/

static void
line_width_rand_seed_value_changed_callback (GtkSpinButton	*widget,
											 PluginVals		*vals)
{/*}}}*/
	vals->line_width_rand_seed = gtk_spin_button_get_value_as_int (widget);
	gtk_widget_queue_draw (widgets.preview);
}/*}}}*/

static void
line_in_margin_value_changed_callback (GtkSpinButton	*widget,
									   PluginVals		*vals)
{/*}}}*/
	vals->line_in_margin = gtk_spin_button_get_value (widget);
	gtk_widget_queue_draw (widgets.preview);
}/*}}}*/

static void
line_in_margin_rand_value_changed_callback (GtkSpinButton	*widget,
											PluginVals		*vals)
{/*}}}*/
	vals->line_in_margin_rand = gtk_spin_button_get_value (widget);
	gtk_widget_queue_draw (widgets.preview);
}/*}}}*/

static void
line_in_margin_rand_seed_value_changed_callback (GtkSpinButton	*widget,
												 PluginVals		*vals)
{/*}}}*/
	vals->line_in_margin_rand_seed = gtk_spin_button_get_value_as_int (widget);
	gtk_widget_queue_draw (widgets.preview);
}/*}}}*/

static void
line_in_length_value_changed_callback (GtkSpinButton	*widget,
									   PluginVals		*vals)
{/*}}}*/
	vals->line_in_length = gtk_spin_button_get_value (widget);
	gtk_widget_queue_draw (widgets.preview);
}/*}}}*/

static void
line_in_length_rand_value_changed_callback (GtkSpinButton	*widget,
											PluginVals		*vals)
{/*}}}*/
	vals->line_in_length_rand = gtk_spin_button_get_value (widget);
	gtk_widget_queue_draw (widgets.preview);
}/*}}}*/

static void
line_in_length_rand_seed_value_changed_callback (GtkSpinButton	*widget,
												 PluginVals		*vals)
{/*}}}*/
	vals->line_in_length_rand_seed = gtk_spin_button_get_value_as_int (widget);
	gtk_widget_queue_draw (widgets.preview);
}/*}}}*/

static void
line_out_margin_value_changed_callback (GtkSpinButton	*widget,
										PluginVals		*vals)
{/*}}}*/
	vals->line_out_margin = gtk_spin_button_get_value (widget);
	gtk_widget_queue_draw (widgets.preview);
}/*}}}*/

static void
line_out_margin_rand_value_changed_callback (GtkSpinButton	*widget,
											 PluginVals		*vals)
{/*}}}*/
	vals->line_out_margin_rand = gtk_spin_button_get_value (widget);
	gtk_widget_queue_draw (widgets.preview);
}/*}}}*/

static void
line_out_margin_rand_seed_value_changed_callback (GtkSpinButton	*widget,
												  PluginVals	*vals)
{/*}}}*/
	vals->line_out_margin_rand_seed = gtk_spin_button_get_value_as_int (widget);
	gtk_widget_queue_draw (widgets.preview);
}/*}}}*/

static void
line_out_length_value_changed_callback (GtkSpinButton	*widget,
										PluginVals		*vals)
{/*}}}*/
	vals->line_out_length = gtk_spin_button_get_value (widget);
	gtk_widget_queue_draw (widgets.preview);
}/*}}}*/

static void
line_out_length_rand_value_changed_callback (GtkSpinButton	*widget,
											 PluginVals		*vals)
{/*}}}*/
	vals->line_out_length_rand = gtk_spin_button_get_value (widget);
	gtk_widget_queue_draw (widgets.preview);
}/*}}}*/

static void
line_out_length_rand_seed_value_changed_callback (GtkSpinButton	*widget,
												  PluginVals	*vals)
{/*}}}*/
	vals->line_out_length_rand_seed = gtk_spin_button_get_value_as_int (widget);
	gtk_widget_queue_draw (widgets.preview);
}/*}}}*/

static void 
freehand_check_button_toggled_callback (GtkToggleButton *widget,
										PluginVals		*vals)
{/*}}}*/
	GtkWidget *check_button = NULL;
	GtkWidget *edit_button = NULL;
	GtkWidget *reset_button = NULL;
	gboolean active;

	active = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget));

	if (GTK_WIDGET (widget) == widgets.radiation_in_freehand) {
		check_button = GTK_WIDGET (widget);
		edit_button =widgets.radiation_in_freehand_edit_button;
		reset_button = widgets.radiation_in_freehand_reset_button;
		vals->radiation_in_freehand_active = active;
	}
	if (GTK_WIDGET (widget) == widgets.radiation_out_freehand) {
		check_button = GTK_WIDGET (widget);
		edit_button =widgets.radiation_out_freehand_edit_button;
		reset_button = widgets.radiation_out_freehand_reset_button;
		vals->radiation_out_freehand_active = active;
	}
	if (GTK_WIDGET (widget) == widgets.parallel_in_freehand) {
		check_button = GTK_WIDGET (widget);
		edit_button =widgets.parallel_in_freehand_edit_button;
		reset_button = widgets.parallel_in_freehand_reset_button;
		vals->parallel_in_freehand_active = active;
	}
	if (GTK_WIDGET (widget) == widgets.parallel_out_freehand) {
		check_button = GTK_WIDGET (widget);
		edit_button =widgets.parallel_out_freehand_edit_button;
		reset_button = widgets.parallel_out_freehand_reset_button;
		vals->parallel_out_freehand_active = active;
	}

	gtk_widget_set_sensitive ( GTK_WIDGET (edit_button), active);
	gtk_widget_set_sensitive ( GTK_WIDGET (reset_button), active);
	if (active == FALSE) {
		preview_freehand_edit_mode_set (FREEHAND_EDIT_MODE_NONE);
		gtk_button_set_label (GTK_BUTTON (edit_button), _("Edit Start"));
	}

	gtk_widget_queue_draw (widgets.preview);
}/*}}}*/

static void
freehand_edit_button_clicked (GtkButton		*widget,
							  PluginVals	*vals)
{/*}}}*/
	FreehandEditModeType type = preview_freehand_edit_mode_get ();
	gtk_button_set_label (GTK_BUTTON (widgets.radiation_in_freehand_edit_button), _("Edit Start"));
	gtk_button_set_label (GTK_BUTTON (widgets.radiation_out_freehand_edit_button), _("Edit Start"));
	gtk_button_set_label (GTK_BUTTON (widgets.parallel_in_freehand_edit_button), _("Edit Start"));
	gtk_button_set_label (GTK_BUTTON (widgets.parallel_out_freehand_edit_button), _("Edit Start"));
	preview_freehand_edit_mode_set (FREEHAND_EDIT_MODE_NONE);

	if (GTK_WIDGET (widget) == widgets.radiation_in_freehand_edit_button &&
		type != FREEHAND_EDIT_MODE_RADIATION_IN) {
		preview_freehand_edit_mode_set (FREEHAND_EDIT_MODE_RADIATION_IN);
		gtk_button_set_label (GTK_BUTTON (widget), _("Edit Stop"));
	}
	if (GTK_WIDGET (widget) == widgets.radiation_out_freehand_edit_button &&
		type != FREEHAND_EDIT_MODE_RADIATION_OUT) {
		preview_freehand_edit_mode_set (FREEHAND_EDIT_MODE_RADIATION_OUT);
		gtk_button_set_label (GTK_BUTTON (widget), _("Edit Stop"));
	}
	if (GTK_WIDGET (widget) == widgets.parallel_in_freehand_edit_button &&
		type != FREEHAND_EDIT_MODE_PARALLEL_IN) {
		preview_freehand_edit_mode_set (FREEHAND_EDIT_MODE_PARALLEL_IN);
		gtk_button_set_label (GTK_BUTTON (widget), _("Edit Stop"));
	}
	if (GTK_WIDGET (widget) == widgets.parallel_out_freehand_edit_button &&
		type != FREEHAND_EDIT_MODE_PARALLEL_OUT) {
		preview_freehand_edit_mode_set (FREEHAND_EDIT_MODE_PARALLEL_OUT);
		gtk_button_set_label (GTK_BUTTON (widget), _("Edit Stop"));
	}
	gtk_widget_queue_draw (widgets.preview);
}/*}}}*/

static void
freehand_reset_button_clicked (GtkButton	*widget,
							   PluginVals	*vals)
{/*}}}*/
	if (GTK_WIDGET (widget) == widgets.radiation_in_freehand_reset_button) {
		radiation_reset_freehand (vals->radiation_in_freehand_points,
								 &vals->radiation_in_freehand_points_count);
	}
	if (GTK_WIDGET (widget) == widgets.radiation_out_freehand_reset_button) {
		radiation_reset_freehand (vals->radiation_out_freehand_points,
								 &vals->radiation_out_freehand_points_count);
	}
	if (GTK_WIDGET (widget) == widgets.parallel_in_freehand_reset_button) {
		parallel_reset_freehand (vals->parallel_in_freehand_points,
								&vals->parallel_in_freehand_points_count);
	}
	if (GTK_WIDGET (widget) == widgets.parallel_out_freehand_reset_button) {
		parallel_reset_freehand (vals->parallel_out_freehand_points,
								&vals->parallel_out_freehand_points_count);
	}
	gtk_widget_queue_draw (widgets.preview);
}/*}}}*/

static void
preset_save (gchar		*path,
			 PluginVals	*vals)
{/*{{{*/
	FILE *fp;
	GKeyFile *keyfile;
	gsize len;
	gchar *data;

	keyfile = pluginvals_to_keyfile (vals);
	data = g_key_file_to_data (keyfile, &len, NULL);

#ifdef _WIN32
    {
        char *mode;
        char *temp;
        mode = g_utf8_to_utf16("wb", -1, NULL, NULL, NULL);
        temp = g_utf8_to_utf16(path, -1, NULL, NULL, NULL);
        fp = _wfopen(temp, mode);
        g_free(mode); mode = NULL;
        g_free(temp); temp = NULL;
    }
#else
	fp = fopen (path, "wb");
#endif
	fprintf (fp, data);
	fclose (fp);

	if (data) g_free (data);
	if (keyfile) g_key_file_free (keyfile);
}/*}}}*/

static void
preset_load (gchar *path)
{/*{{{*/

	GKeyFile *keyfile;
	GError *error;

	keyfile = g_key_file_new ();
	if (g_key_file_load_from_file (keyfile, path, G_KEY_FILE_KEEP_COMMENTS, &error) == FALSE) {
		msgbox (GTK_MESSAGE_ERROR, GTK_BUTTONS_CLOSE, NULL, error->message);
		return;
	}
	
	keyfile_to_widgets (keyfile);
	g_key_file_free (keyfile);
}/*}}}*/

static void
preset_delete_callback (GtkWidget    *dummy,
						gpointer     data)
{/*{{{*/
	GtkResponseType ret;
	gchar *fname, *path;
	gchar msg[1024];
	gboolean has_sel;
	GtkTreeIter iter;
	GtkTreeModel *model;
	GtkTreeSelection *sel;

	if (preset_dir == NULL) {
		preset_dir = get_preset_dir ();
		if (preset_dir == NULL) return;
	}
    //FIXME: Preset̐ݒ\c[̃EBWFbgZbg
    sel = gtk_tree_view_get_selection (
            GTK_TREE_VIEW (widgets.preset_tree));
    has_sel = gtk_tree_selection_get_selected (sel, &model, &iter);
    if (has_sel == FALSE) return;

    gtk_tree_model_get (model, &iter, 0, &fname, -1);

    g_snprintf (msg, sizeof (msg), _("Do you want to delete '%s' ?"), fname);
    ret = msgbox (
            GTK_MESSAGE_QUESTION,
            GTK_BUTTONS_OK_CANCEL, 
            widgets.dialog, msg);
    if (ret != GTK_RESPONSE_OK) return;

    path = g_build_filename (preset_dir, fname, NULL);
    g_free (fname);
    if (g_file_test (path, G_FILE_TEST_EXISTS)) {
        if (g_remove (path) == 0)
            gtk_list_store_remove (GTK_LIST_STORE (model), &iter);
    }
    g_free (path);
}/*}}}*/

static void
preset_update_callback (GtkWidget    *dummy,
						gpointer     data)
{/*{{{*/
	GDir *dir;
	const gchar *fname;

    gtk_list_store_clear (widgets.preset_store);

	if (preset_dir == NULL) {
		preset_dir = get_preset_dir ();
		if (preset_dir == NULL) return;
	}
	dir = g_dir_open (preset_dir, 0, NULL);
	if (dir == NULL) return;

	while ((fname = g_dir_read_name (dir))) {
		gchar *fpath = g_build_filename (preset_dir, fname, NULL);

		if (g_file_test (fpath, G_FILE_TEST_IS_DIR) == FALSE) {
			GtkTreeIter iter;
			gtk_list_store_append (widgets.preset_store, &iter);
			gtk_list_store_set (widgets.preset_store, &iter, 0, fname, -1);
		}
		g_free (fpath);
	}
	g_dir_close (dir); 
}/*}}}*/

static void
preset_load_callback (GtkButton  *dummy,
					  gpointer   data)
{/*{{{*/
	gchar *fname, *path;
	gboolean has_sel;
	GtkTreeIter iter;
	GtkTreeModel *model;
	GtkTreeSelection *sel;

	if (preset_dir == NULL) {
		preset_dir = get_preset_dir ();
		if (preset_dir == NULL) return;
	}

	sel = gtk_tree_view_get_selection ( GTK_TREE_VIEW (widgets.preset_tree));
	has_sel = gtk_tree_selection_get_selected (sel, &model, &iter);
	if (has_sel == FALSE) return;

	gtk_tree_model_get (model, &iter, 0, &fname, -1);
	path = g_build_filename (preset_dir, fname, NULL);

	preset_load (path);
	goto done;

done:
	preset_update_callback (NULL, NULL);
	if (fname) g_free (fname);
	if (path) g_free (path);
	return;
}/*}}}*/

static void
preset_save_ok_callback (GtkWidget   *dummy,
						 const gchar *name,
						 gpointer    data)
{/*{{{*/
	PluginVals *vals;
	gchar *path = NULL;
	gchar *name_locale = NULL;
	GtkResponseType ret;
	gchar msg[2048];

	vals = (PluginVals*)data;

	if (preset_dir == NULL) {
		preset_dir = get_preset_dir ();
		if (preset_dir == NULL) return;
	}

	name_locale = g_filename_from_utf8 (name, -1, NULL, NULL, NULL);

	if (name_locale) {
		path = g_build_filename (preset_dir, name_locale, NULL);
		g_free (name_locale);
		name_locale = NULL;
	}

	if (g_file_test (path, G_FILE_TEST_EXISTS) == TRUE) {
		g_snprintf (msg, sizeof (msg),
				_("The name '%s' was already exists. Overwrite?"), name);
		ret = msgbox (
				GTK_MESSAGE_QUESTION,
				GTK_BUTTONS_OK_CANCEL, 
				widgets.dialog, msg);
		if (ret != GTK_RESPONSE_OK) goto done;
	}

	preset_save (path, vals);

done:
	preset_update_callback (NULL, NULL);
	if (path) g_free (path);
	return;
}/*}}}*/

static void
preset_save_callback (GtkWidget  *dummy,
					  gpointer   data)
{/*{{{*/
	GtkWidget *dialog;
	PluginVals *vals;

	vals = (PluginVals*)data;

	dialog = gimp_query_string_box (_("New Parameter Preset"), 
		widgets.dialog, 
		 gimp_standard_help_func, 0, 
		_("Enter a name for the new Parameter Preset"), 
		_("Unnamed"), 
		NULL, NULL,
		preset_save_ok_callback, vals);
	gtk_window_set_modal (GTK_WINDOW (dialog), TRUE);
	gtk_window_set_keep_above (GTK_WINDOW (dialog), TRUE);
	gtk_widget_show (dialog);
}/*}}}*/

static void
preset_tree_dbclicked (GtkTreeView   *tree_view,
					   GtkTreeIter   *iter,
					   GtkTreePath   *path,
					   gpointer      user_data)
{/*{{{*/
	int i;
	GtkTreeSelection *sel;
	GtkTreeModel *model;
	GtkTreeIter siter;
	GtkTreePath *spath;
	gboolean has_sel;
	
	model = gtk_tree_view_get_model (tree_view);
	sel = gtk_tree_view_get_selection (tree_view);
	has_sel = gtk_tree_selection_get_selected (
			sel, &model, &siter);
	if (!has_sel) return;
	spath = gtk_tree_model_get_path (model, &siter);

	preset_load_callback (NULL, NULL);
	
	//_ł
	for (i=0; i< 5; i++) {
		if ((i%2) == 0) {
			gtk_tree_selection_select_path (sel, spath);
		} else {
			gtk_tree_selection_unselect_path (sel, spath);
		}
		while (gtk_events_pending ()) {
			gtk_main_iteration ();
		}
		g_usleep (50000);
	}
	gtk_tree_selection_select_path (sel, spath);
	gtk_tree_path_free (spath);
}/*}}}*/

static GtkWidget*
build_gtk_spin_button_new (double min,
						   double max,
						   double step,
						   guint digits,
						   double value)
{
	GtkWidget *spin = gtk_spin_button_new_with_range (min, max, step);
	gtk_spin_button_set_digits (GTK_SPIN_BUTTON (spin), digits);
	gtk_spin_button_set_value (GTK_SPIN_BUTTON (spin), value);
	gtk_entry_set_alignment (GTK_ENTRY (spin), 1.0);
	return spin;
}

static GtkWidget*
build_line_antialias_combo_box_new (AntialiasType default_value)
{
	GtkWidget *combo = gimp_int_combo_box_new (
		_("None"), ANTIALIAS_TYPE_NONE,
		_("Fast"), ANTIALIAS_TYPE_FAST,
		_("Best"), ANTIALIAS_TYPE_BEST,
		NULL
		);
	gimp_int_combo_box_set_active (GIMP_INT_COMBO_BOX (combo), default_value);
	return combo;
}

static GtkWidget*
build_line_pitch_type_combo_box_new (LinePitchType default_value)
{
	GtkWidget *combo = gimp_int_combo_box_new (
		_("IN"),  LINE_PITCH_TYPE_IN,
		_("OUT"), LINE_PITCH_TYPE_OUT,
		NULL
		);
	gimp_int_combo_box_set_active (GIMP_INT_COMBO_BOX (combo), default_value);
	return combo;
}

static GtkWidget*
build_line_wave_type_combo_box_new (LineWaveType default_value)
{
	GtkWidget *combo = gimp_int_combo_box_new (
		_("Default"),   LINE_WAVE_TYPE_DEFAULT,
		_("Cos"),       LINE_WAVE_TYPE_COS,
		_("Arc"),       LINE_WAVE_TYPE_ARC,
		_("Arc(Inv)"),  LINE_WAVE_TYPE_ARC_INVERT,
		_("Block"),     LINE_WAVE_TYPE_BLOCK,
		_("Block(Inv)"),LINE_WAVE_TYPE_BLOCK_INVERT,
		_("N"),         LINE_WAVE_TYPE_N,
		NULL
		);
	gimp_int_combo_box_set_active (GIMP_INT_COMBO_BOX (combo), default_value);
	return combo;
}

static GtkWidget*
build_unit_combo_box_new (gboolean has_percent,
						  GimpUnit default_value)
{
	GtkWidget *combo;

	combo = gimp_int_combo_box_new (
		gimp_unit_get_plural (GIMP_UNIT_PIXEL), GIMP_UNIT_PIXEL,
		gimp_unit_get_plural (GIMP_UNIT_INCH),  GIMP_UNIT_INCH,
		gimp_unit_get_plural (GIMP_UNIT_MM),    GIMP_UNIT_MM,
		gimp_unit_get_plural (GIMP_UNIT_POINT), GIMP_UNIT_POINT,
		gimp_unit_get_plural (GIMP_UNIT_PICA),  GIMP_UNIT_PICA,
		NULL);

	if (has_percent) {
		gimp_int_combo_box_append (GIMP_INT_COMBO_BOX (combo),
			GIMP_INT_STORE_VALUE, GIMP_UNIT_PERCENT,
			GIMP_INT_STORE_LABEL, gimp_unit_get_plural (GIMP_UNIT_PERCENT),
			-1
		);
	}

	gimp_int_combo_box_set_active (GIMP_INT_COMBO_BOX (combo), default_value);
	return combo;
}

static GtkWidget*
build_freehand_widgets (GtkWidget **check_button,
						GtkWidget **edit_button,
						GtkWidget **reset_button,
						PluginVals *vals)
{
	GtkWidget *box = gtk_box_new (GTK_ORIENTATION_HORIZONTAL, 0);
	*check_button = gtk_check_button_new_with_label (_("FreeHand"));
	*edit_button = gtk_button_new_with_label (_("Edit Start"));
	*reset_button = gtk_button_new_with_label (_("Reset"));
	gtk_widget_set_sensitive ( GTK_WIDGET (*edit_button), FALSE);
	gtk_widget_set_sensitive ( GTK_WIDGET (*reset_button), FALSE);

	g_signal_connect (*check_button, "toggled", G_CALLBACK (freehand_check_button_toggled_callback), vals);
	g_signal_connect (*edit_button, "clicked", G_CALLBACK (freehand_edit_button_clicked), vals);
	g_signal_connect (*reset_button, "clicked", G_CALLBACK (freehand_reset_button_clicked), vals);

	gtk_box_pack_start(GTK_BOX (box), *check_button, FALSE, FALSE, 0);
	gtk_box_pack_start(GTK_BOX (box), *edit_button,  FALSE, FALSE, 0);
	gtk_box_pack_start(GTK_BOX (box), *reset_button, FALSE, FALSE, 0);
	gtk_box_pack_start(GTK_BOX (box), gtk_label_new(""), TRUE, TRUE, 0);
	return box;
}


static int
run_dialog (PluginVals *vals)
{/*{{{*/
#define PADDING 4

/* tree of containers
main_hbox
+-widgets.preview
+-preview_controller_vbox
+-base_notebook
  +-options_vbox
  | +-area_option_frame
  | | +-area_option_vbox
  | |   +-area_option_table
  | |   +-(gtk_separator)
  | |   +-area_type_notebook
  | +-line_option_frame
  |   +-line_option_table
  +-preset_table
    +-widgets.preset_frame
*/

	GtkWidget *main_hbox;
	GtkWidget *preview_controller_vbox;
	GtkWidget *base_notebook;
	GtkWidget *options_vbox;
	GtkWidget *preset_table;

	gimp_ui_init (PLUGIN_NAME, TRUE);
	
	//show_splash ();

	// create dialog
	widgets.dialog = gtk_dialog_new ();
	widgets.topmost = gtk_check_button_new_with_label (_("Topmost"));
	gtk_window_set_title (GTK_WINDOW (widgets.dialog), 
			"Speedline" " Version " PLUGIN_VERSION);

	gtk_window_set_icon (GTK_WINDOW (widgets.dialog), 
	gdk_pixbuf_new_from_inline (-1, img_window_icon, FALSE, NULL));
	gtk_window_set_role (GTK_WINDOW (widgets.dialog), PLUGIN_NAME);
	gtk_window_set_resizable (GTK_WINDOW (widgets.dialog), FALSE);
	gtk_dialog_set_has_separator (GTK_DIALOG (widgets.dialog), FALSE);
	gtk_dialog_add_buttons (GTK_DIALOG (widgets.dialog),
#ifdef _WIN32
		//_("Modify"),    SPEEDLINE_RESPONSE_MODIFY,
		_("Reset"),     SPEEDLINE_RESPONSE_RESET,
		_("Create"),    SPEEDLINE_RESPONSE_CREATE,
		_("Close"),     SPEEDLINE_RESPONSE_CLOSE,
#else
		_("Reset"),     SPEEDLINE_RESPONSE_RESET,
		_("Close"),     SPEEDLINE_RESPONSE_CLOSE,
		_("Create"),    SPEEDLINE_RESPONSE_CREATE,
		//_("Modify"),    SPEEDLINE_RESPONSE_MODIFY,
#endif
		NULL);
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widgets.topmost), vals->gui.topmost);

	gtk_dialog_add_action_widget (GTK_DIALOG (widgets.dialog), widgets.topmost, SPEEDLINE_RESPONSE_TOPMOST);
	gtk_button_box_set_child_secondary (GTK_BUTTON_BOX (GTK_DIALOG (widgets.dialog)->action_area), widgets.topmost, TRUE);
	
	g_signal_connect (widgets.dialog, "response", G_CALLBACK (response_callback), vals);
	g_signal_connect (widgets.dialog, "destroy",  G_CALLBACK (gtk_main_quit), NULL);
	
	// main hbox
	main_hbox = gtk_hbox_new (FALSE, 2);
	gtk_container_set_border_width (GTK_CONTAINER (main_hbox), 2);
	gtk_container_add (GTK_CONTAINER (GTK_DIALOG (widgets.dialog)->vbox ), main_hbox);

    //preview
    widgets.preview = preview_new (vals);
	gtk_box_pack_start (GTK_BOX (main_hbox), GTK_WIDGET (widgets.preview), FALSE, FALSE, 0);

	//preview controller
	preview_controller_vbox = gtk_vbox_new (FALSE, 2);
	gtk_box_pack_start (GTK_BOX (main_hbox), GTK_WIDGET (preview_controller_vbox), FALSE, FALSE, 0);

	widgets.preview_plus_button = gtk_button_new ();
	gtk_button_set_label (GTK_BUTTON (widgets.preview_plus_button), "+");
	gtk_widget_set_size_request (GTK_WIDGET (widgets.preview_plus_button), 24, 24);
	gtk_box_pack_start (GTK_BOX (preview_controller_vbox), GTK_WIDGET (widgets.preview_plus_button), FALSE, FALSE, 0);
	g_signal_connect (GTK_WIDGET (widgets.preview_plus_button), "clicked", G_CALLBACK (preview_controller_button_clicked_callback), vals);

	widgets.preview_minus_button = gtk_button_new ();
	gtk_button_set_label (GTK_BUTTON (widgets.preview_minus_button), "-");
	gtk_widget_set_size_request (GTK_WIDGET (widgets.preview_minus_button), 24, 24);
	gtk_box_pack_start (GTK_BOX (preview_controller_vbox), GTK_WIDGET (widgets.preview_minus_button), FALSE, FALSE, 0);
	g_signal_connect (GTK_WIDGET (widgets.preview_minus_button), "clicked", G_CALLBACK (preview_controller_button_clicked_callback), vals);
	gtk_widget_set_sensitive (GTK_WIDGET (widgets.preview_plus_button), FALSE);


	base_notebook = gtk_notebook_new ();

	gtk_box_pack_start (GTK_BOX (main_hbox), GTK_WIDGET (base_notebook), FALSE, FALSE, PADDING);
	gtk_notebook_set_show_tabs (GTK_NOTEBOOK (base_notebook), TRUE);
	gtk_notebook_set_show_border (GTK_NOTEBOOK (base_notebook), TRUE);
	g_signal_connect (base_notebook, "switch_page", G_CALLBACK (basenotebook_callback), vals);

	// Page1
	options_vbox = gtk_vbox_new (FALSE, 2);
	gtk_container_set_border_width (GTK_CONTAINER (options_vbox), PADDING);
	gtk_notebook_append_page (GTK_NOTEBOOK (base_notebook), GTK_WIDGET (options_vbox), gtk_label_new (_("Options")));

	// Page2
	preset_table = gtk_table_new (5, 2, FALSE);
	gtk_container_set_border_width (GTK_CONTAINER (preset_table), PADDING);
	gtk_notebook_append_page (GTK_NOTEBOOK (base_notebook), GTK_WIDGET (preset_table), gtk_label_new (_("Preset")));

	//area 
	{
		GtkWidget *area_option_frame;
		GtkWidget *area_option_vbox;

		GtkWidget *area_option_table;

		GtkWidget *area_type_note_radiation_table = gtk_table_new (2, 5, FALSE);
		GtkWidget *area_type_note_parallel_table = gtk_table_new (3, 7, FALSE);

		//frame
		area_option_frame = gtk_frame_new (_("Area Option"));
		area_option_vbox = gtk_vbox_new (FALSE, 0);
		gtk_container_add (GTK_CONTAINER (area_option_frame), area_option_vbox);
		gtk_box_pack_start (GTK_BOX (options_vbox), GTK_WIDGET (area_option_frame), FALSE, TRUE, PADDING);

		area_option_table = gtk_table_new (1, 6, FALSE);
		gtk_box_pack_start (GTK_BOX (area_option_vbox), GTK_WIDGET (area_option_table), FALSE, FALSE, PADDING);
		

		//area type
		{
			GtkWidget *label = gtk_label_new (_("Type:"));
			gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
			gtk_table_attach (GTK_TABLE (area_option_table), label, 0, 1, 0, 1, GTK_FILL, GTK_FILL, PADDING, PADDING);
		}

		
		widgets.area_type_radiation = gtk_radio_button_new_with_label (NULL, _("Radiation"));
		gtk_table_attach (GTK_TABLE (area_option_table), GTK_WIDGET (widgets.area_type_radiation), 1, 2, 0, 1, 0, 0, 0, PADDING);

		widgets.area_type_parallel = gtk_radio_button_new_with_label (
			gtk_radio_button_get_group (GTK_RADIO_BUTTON (widgets.area_type_radiation)), _("Parallel"));
		gtk_table_attach (GTK_TABLE (area_option_table), GTK_WIDGET (widgets.area_type_parallel), 2, 3, 0, 1, 0, 0, 0, PADDING);

		//callback for radio buttons
		g_signal_connect (GTK_WIDGET (widgets.area_type_radiation), "clicked", G_CALLBACK (area_type_radio_click_callback), vals);
		g_signal_connect (GTK_WIDGET (widgets.area_type_parallel), "clicked", G_CALLBACK (area_type_radio_click_callback), vals);
		{
			GtkWidget *temp = NULL;
			switch (vals->area_type) {
				case AREA_TYPE_RADIATION: temp = widgets.area_type_radiation; break;
				case AREA_TYPE_PARALLEL:  temp = widgets.area_type_parallel; break;
			}
			gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (temp), TRUE);
		}

		//space
		{
			GtkWidget *label = gtk_label_new (_("       "));
			gtk_table_attach (GTK_TABLE (area_option_table), label, 3, 4, 0, 1, GTK_FILL, GTK_FILL, PADDING, PADDING);
		}

		//area type unit
		{
			GtkWidget *label = gtk_label_new (_("Unit:"));
			gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
			gtk_table_attach (GTK_TABLE (area_option_table), label, 4, 5, 0, 1, GTK_FILL, GTK_FILL, PADDING, PADDING);
		}
		widgets.area_type_unit = build_unit_combo_box_new (FALSE, vals->area_unit);
		g_signal_connect (GTK_WIDGET (widgets.area_type_unit), "changed", G_CALLBACK (area_type_unit_changed_callback), vals);
		gtk_table_attach (GTK_TABLE (area_option_table), widgets.area_type_unit, 5, 6, 0, 1, 0, 0, PADDING, PADDING);

		// area type details
		widgets.area_type_notebook = gtk_notebook_new ();
		gtk_notebook_set_show_tabs (GTK_NOTEBOOK (widgets.area_type_notebook), FALSE);
		gtk_notebook_set_show_border (GTK_NOTEBOOK (widgets.area_type_notebook), FALSE);

		gtk_notebook_append_page (GTK_NOTEBOOK (widgets.area_type_notebook), area_type_note_radiation_table, gtk_label_new (_("Radiation")));
		gtk_notebook_append_page (GTK_NOTEBOOK (widgets.area_type_notebook), area_type_note_parallel_table, gtk_label_new (_("Parallel")));

		gtk_box_pack_start (GTK_BOX (area_option_vbox), gtk_separator_new (GTK_ORIENTATION_HORIZONTAL), FALSE, FALSE, PADDING);
		gtk_box_pack_start (GTK_BOX (area_option_vbox), GTK_WIDGET (widgets.area_type_notebook), FALSE, FALSE, 0);


		//radiation area options
		{
			GtkWidget *label_center     = gtk_label_new (_("Center:"));
			GtkWidget *label_center_x   = gtk_label_new (_("X:"));
			GtkWidget *label_center_y   = gtk_label_new (_("Y:"));
			GtkWidget *label_oblateness = gtk_label_new (_("Oblateness:"));
			GtkWidget *label_radius     = gtk_label_new (_("Radius:"));
			GtkWidget *label_radius_in  = gtk_label_new (_("In:"));
			GtkWidget *label_radius_out = gtk_label_new (_("Out:"));
			GtkWidget *radiation_in_freehand_box = build_freehand_widgets (&widgets.radiation_in_freehand,
																		   &widgets.radiation_in_freehand_edit_button,
																		   &widgets.radiation_in_freehand_reset_button,
																		   vals);
			GtkWidget *radiation_out_freehand_box = build_freehand_widgets (&widgets.radiation_out_freehand,
																			&widgets.radiation_out_freehand_edit_button,
																			&widgets.radiation_out_freehand_reset_button,
																			vals);

			gtk_misc_set_alignment (GTK_MISC (label_center),     0.0, 0.5);
			gtk_misc_set_alignment (GTK_MISC (label_center_x),   0.0, 0.5);
			gtk_misc_set_alignment (GTK_MISC (label_center_y),   0.0, 0.5);
			gtk_misc_set_alignment (GTK_MISC (label_oblateness), 0.0, 0.5);
			gtk_misc_set_alignment (GTK_MISC (label_radius),     0.0, 0.5);
			gtk_misc_set_alignment (GTK_MISC (label_radius_in),  0.0, 0.5);
			gtk_misc_set_alignment (GTK_MISC (label_radius_out), 0.0, 0.5);

			widgets.radiation_x = build_gtk_spin_button_new (SHRT_MIN, SHRT_MAX, 1, 3, vals->radiation_x);
			widgets.radiation_y = build_gtk_spin_button_new (SHRT_MIN, SHRT_MAX, 1, 3, vals->radiation_y);
			widgets.radiation_oblateness = build_gtk_spin_button_new (-4.00, 0.95, 0.01, 3, vals->radiation_oblateness);
			widgets.radiation_in_radius  = build_gtk_spin_button_new (0, SHRT_MAX, 1, 3, vals->radiation_in_radius);
			widgets.radiation_out_radius = build_gtk_spin_button_new (0, SHRT_MAX, 1, 3, vals->radiation_out_radius);

			g_signal_connect (widgets.radiation_x, "value_changed", G_CALLBACK (radiation_x_value_changed_callback), vals);
			g_signal_connect (widgets.radiation_y, "value_changed", G_CALLBACK (radiation_y_value_changed_callback), vals);
			g_signal_connect (widgets.radiation_oblateness, "value_changed", G_CALLBACK (radiation_oblateness_value_changed_callback), vals);
			g_signal_connect (widgets.radiation_in_radius, "value_changed", G_CALLBACK (radiation_in_radius_value_changed_callback), vals);
			g_signal_connect (widgets.radiation_out_radius, "value_changed", G_CALLBACK (radiation_out_radius_value_changed_callback), vals);

			gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widgets.radiation_in_freehand), vals->radiation_in_freehand_active);
			gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widgets.radiation_out_freehand), vals->radiation_out_freehand_active);

			gtk_table_attach (GTK_TABLE (area_type_note_radiation_table), label_center,                 0, 1, 0, 1, 0, 0, PADDING, PADDING);
			gtk_table_attach (GTK_TABLE (area_type_note_radiation_table), label_center_x,               1, 2, 0, 1, 0, 0, PADDING, PADDING);
			gtk_table_attach (GTK_TABLE (area_type_note_radiation_table), widgets.radiation_x,          2, 3, 0, 1, 0, 0, PADDING, PADDING);
			gtk_table_attach (GTK_TABLE (area_type_note_radiation_table), label_center_y,               3, 4, 0, 1, 0, 0, PADDING, PADDING);
			gtk_table_attach (GTK_TABLE (area_type_note_radiation_table), widgets.radiation_y,          4, 5, 0, 1, 0, 0, PADDING, PADDING);
			gtk_table_attach (GTK_TABLE (area_type_note_radiation_table), label_oblateness,             5, 6, 0, 1, 0, 0, PADDING, PADDING);
			gtk_table_attach (GTK_TABLE (area_type_note_radiation_table), widgets.radiation_oblateness, 6, 7, 0, 1, 0, 0, PADDING, PADDING);

			gtk_table_attach (GTK_TABLE (area_type_note_radiation_table), label_radius,                 0, 1, 1, 2, 0, 0, PADDING, PADDING);
			gtk_table_attach (GTK_TABLE (area_type_note_radiation_table), label_radius_in,              1, 2, 1, 2, 0, 0, PADDING, PADDING);
			gtk_table_attach (GTK_TABLE (area_type_note_radiation_table), widgets.radiation_in_radius,  2, 3, 1, 2, 0, 0, PADDING, PADDING);
			gtk_table_attach (GTK_TABLE (area_type_note_radiation_table), radiation_in_freehand_box,    3, 6, 1, 2, 0, 0, PADDING, PADDING);

			gtk_table_attach (GTK_TABLE (area_type_note_radiation_table), label_radius_out,             1, 2, 2, 3, 0, 0, PADDING, PADDING);
			gtk_table_attach (GTK_TABLE (area_type_note_radiation_table), widgets.radiation_out_radius, 2, 3, 2, 3, 0, 0, PADDING, PADDING);
			gtk_table_attach (GTK_TABLE (area_type_note_radiation_table), radiation_out_freehand_box,   3, 6, 2, 3, 0, 0, PADDING, PADDING);
		}


		//parallel area options
		{
			GtkWidget *label_in        = gtk_label_new (_("In:"));
			GtkWidget *label_in_x      = gtk_label_new (_("X:"));
			GtkWidget *label_in_y      = gtk_label_new (_("Y:"));
			GtkWidget *label_in_range  = gtk_label_new (_("Range:"));
			GtkWidget *label_out       = gtk_label_new (_("Out:"));
			GtkWidget *label_out_x     = gtk_label_new (_("X:"));
			GtkWidget *label_out_y     = gtk_label_new (_("Y:"));
			GtkWidget *label_out_range = gtk_label_new (_("Range:"));
			GtkWidget *parallel_in_freehand_box = build_freehand_widgets (&widgets.parallel_in_freehand,
																		  &widgets.parallel_in_freehand_edit_button,
																		  &widgets.parallel_in_freehand_reset_button,
																		  vals);
			GtkWidget *parallel_out_freehand_box = build_freehand_widgets (&widgets.parallel_out_freehand,
																		   &widgets.parallel_out_freehand_edit_button,
																		   &widgets.parallel_out_freehand_reset_button,
																		   vals);

			gtk_misc_set_alignment (GTK_MISC (label_in),        0.0, 0.5);
			gtk_misc_set_alignment (GTK_MISC (label_in_x),      0.0, 0.5);
			gtk_misc_set_alignment (GTK_MISC (label_in_y),      0.0, 0.5);
			gtk_misc_set_alignment (GTK_MISC (label_in_range),  0.0, 0.5);
			gtk_misc_set_alignment (GTK_MISC (label_out),       0.0, 0.5);
			gtk_misc_set_alignment (GTK_MISC (label_out_x),     0.0, 0.5);
			gtk_misc_set_alignment (GTK_MISC (label_out_y),     0.0, 0.5);
			gtk_misc_set_alignment (GTK_MISC (label_out_range), 0.0, 0.5);

			widgets.parallel_in_x = build_gtk_spin_button_new (SHRT_MIN, SHRT_MAX, 1, 3, vals->parallel_in_x);
			widgets.parallel_in_y = build_gtk_spin_button_new (SHRT_MIN, SHRT_MAX, 1, 3, vals->parallel_in_y);
			widgets.parallel_in_range = build_gtk_spin_button_new (0.001, SHRT_MAX, 1, 3, vals->parallel_in_range);
			widgets.parallel_out_x = build_gtk_spin_button_new (SHRT_MIN, SHRT_MAX, 1, 3, vals->parallel_out_x);
			widgets.parallel_out_y = build_gtk_spin_button_new (SHRT_MIN, SHRT_MAX, 1, 3, vals->parallel_out_y);
			widgets.parallel_out_range = build_gtk_spin_button_new (0.001, SHRT_MAX, 1, 3, vals->parallel_out_range);

			gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widgets.parallel_out_freehand), vals->parallel_out_freehand_active);
			gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widgets.parallel_in_freehand), vals->parallel_in_freehand_active);

			g_signal_connect (widgets.parallel_in_x, "value_changed", G_CALLBACK (parallel_in_x_value_changed_callback), vals);
			g_signal_connect (widgets.parallel_in_y, "value_changed", G_CALLBACK (parallel_in_y_value_changed_callback), vals);
			g_signal_connect (widgets.parallel_in_range, "value_changed", G_CALLBACK (parallel_in_range_value_changed_callback), vals);
			g_signal_connect (widgets.parallel_out_x, "value_changed", G_CALLBACK (parallel_out_x_value_changed_callback), vals);
			g_signal_connect (widgets.parallel_out_y, "value_changed", G_CALLBACK (parallel_out_y_value_changed_callback), vals);
			g_signal_connect (widgets.parallel_out_range, "value_changed", G_CALLBACK (parallel_out_range_value_changed_callback), vals);

			gtk_table_attach (GTK_TABLE (area_type_note_parallel_table), label_in,                  0, 1, 0, 1, 0, 0, PADDING, PADDING);
			gtk_table_attach (GTK_TABLE (area_type_note_parallel_table), label_in_x,                1, 2, 0, 1, 0, 0, PADDING, PADDING);
			gtk_table_attach (GTK_TABLE (area_type_note_parallel_table), widgets.parallel_in_x,     2, 3, 0, 1, 0, 0, PADDING, PADDING);
			gtk_table_attach (GTK_TABLE (area_type_note_parallel_table), label_in_y,                3, 4, 0, 1, 0, 0, PADDING, PADDING);
			gtk_table_attach (GTK_TABLE (area_type_note_parallel_table), widgets.parallel_in_y,     4, 5, 0, 1, 0, 0, PADDING, PADDING);
			gtk_table_attach (GTK_TABLE (area_type_note_parallel_table), label_in_range,            5, 6, 0, 1, 0, 0, PADDING, PADDING);
			gtk_table_attach (GTK_TABLE (area_type_note_parallel_table), widgets.parallel_in_range, 6, 7, 0, 1, 0, 0, PADDING, PADDING);

			gtk_table_attach (GTK_TABLE (area_type_note_parallel_table), parallel_in_freehand_box,  1, 7, 1, 2, GTK_FILL, 0, PADDING, PADDING);

			gtk_table_attach (GTK_TABLE (area_type_note_parallel_table), label_out,                  0, 1, 2, 3, 0, 0, PADDING, PADDING);
			gtk_table_attach (GTK_TABLE (area_type_note_parallel_table), label_out_x,                1, 2, 2, 3, 0, 0, PADDING, PADDING);
			gtk_table_attach (GTK_TABLE (area_type_note_parallel_table), widgets.parallel_out_x,     2, 3, 2, 3, 0, 0, PADDING, PADDING);
			gtk_table_attach (GTK_TABLE (area_type_note_parallel_table), label_out_y,                3, 4, 2, 3, 0, 0, PADDING, PADDING);
			gtk_table_attach (GTK_TABLE (area_type_note_parallel_table), widgets.parallel_out_y,     4, 5, 2, 3, 0, 0, PADDING, PADDING);
			gtk_table_attach (GTK_TABLE (area_type_note_parallel_table), label_out_range,            5, 6, 2, 3, 0, 0, PADDING, PADDING);
			gtk_table_attach (GTK_TABLE (area_type_note_parallel_table), widgets.parallel_out_range, 6, 7, 2, 3, 0, 0, PADDING, PADDING);

			gtk_table_attach (GTK_TABLE (area_type_note_parallel_table), parallel_out_freehand_box,  1, 7, 3, 4, GTK_FILL, 0, PADDING, PADDING);
		}
	}

	//line option
	{
		GtkWidget *line_option_frame;
		GtkWidget *line_option_table;

		line_option_frame = gtk_frame_new (_("Line Option"));
		line_option_table = gtk_table_new (8, 7, FALSE);
		gtk_container_add (GTK_CONTAINER (line_option_frame), line_option_table);
		gtk_box_pack_start (GTK_BOX (options_vbox), GTK_WIDGET (line_option_frame), FALSE, TRUE, 0);
		{
			GtkWidget *label_color           = gtk_label_new (_("Color:"));
			GtkWidget *label_antialias       = gtk_label_new (_("Antialias:"));
			GtkWidget *label_amount          = gtk_label_new (_("Amount"));
			GtkWidget *label_rand            = gtk_label_new (_("Rand"));
			GtkWidget *label_rand_seed       = gtk_label_new (_("RandSeed"));
			GtkWidget *label_type            = gtk_label_new (_("Type"));
			GtkWidget *label_pitch           = gtk_label_new (_("Pitch:"));
			GtkWidget *label_block_on_pitch  = gtk_label_new (_("On Pitch:"));
			GtkWidget *label_block_off_pitch = gtk_label_new (_("Off Picth:"));
			GtkWidget *label_wave_npoint     = gtk_label_new (_("Wave NPoint:"));
			GtkWidget *label_wave_depth      = gtk_label_new (_("Wave Depth:"));
			GtkWidget *label_width           = gtk_label_new (_("Width:"));
			GtkWidget *label_in_margin       = gtk_label_new (_("In Margin:"));
			GtkWidget *label_in_length       = gtk_label_new (_("In Length:"));
			GtkWidget *label_out_margin      = gtk_label_new (_("Out Margin:"));
			GtkWidget *label_out_length      = gtk_label_new (_("Out Length:"));

			gtk_misc_set_alignment (GTK_MISC (label_color),      0.0, 0.5);
			gtk_misc_set_alignment (GTK_MISC (label_antialias),  0.0, 0.5);
			//gtk_misc_set_alignment (GTK_MISC (label_amount),     0.0, 0.5);
			//gtk_misc_set_alignment (GTK_MISC (label_rand),       0.0, 0.5);
			//gtk_misc_set_alignment (GTK_MISC (label_rand_seed),  0.0, 0.5);
			gtk_misc_set_alignment (GTK_MISC (label_pitch),      0.0, 0.5);
			gtk_misc_set_alignment (GTK_MISC (label_block_on_pitch), 0.0, 0.5);
			gtk_misc_set_alignment (GTK_MISC (label_block_off_pitch), 0.0, 0.5);
			gtk_misc_set_alignment (GTK_MISC (label_wave_npoint), 0.0, 0.5);
			gtk_misc_set_alignment (GTK_MISC (label_wave_depth), 0.0, 0.5);
			gtk_misc_set_alignment (GTK_MISC (label_width),      0.0, 0.5);
			gtk_misc_set_alignment (GTK_MISC (label_in_margin),  0.0, 0.5);
			gtk_misc_set_alignment (GTK_MISC (label_in_length),  0.0, 0.5);
			gtk_misc_set_alignment (GTK_MISC (label_out_margin), 0.0, 0.5);
			gtk_misc_set_alignment (GTK_MISC (label_out_length), 0.0, 0.5);

			widgets.line_color                = gimp_color_button_new (_("Line Color"), 20, 10, &vals->line_color, GIMP_COLOR_AREA_FLAT);
			widgets.line_antialias            = build_line_antialias_combo_box_new (vals->line_antialias);
			widgets.line_pitch                = build_gtk_spin_button_new (0.5, SHRT_MAX, 1, 3, vals->line_pitch);
			widgets.line_pitch_rand           = build_gtk_spin_button_new (0, 100 , 1, 2, vals->line_pitch_rand);
			widgets.line_pitch_rand_seed      = build_gtk_spin_button_new (0, SHRT_MAX, 1, 0, vals->line_pitch_rand_seed);
			widgets.line_pitch_type           = build_line_pitch_type_combo_box_new (vals->line_pitch_type);
			widgets.line_block_on_pitch       = build_gtk_spin_button_new (0, SHRT_MAX, 1, 3, vals->line_block_on_pitch);
			widgets.line_block_on_pitch_rand  = build_gtk_spin_button_new (0, 100 , 1, 2, vals->line_block_on_pitch_rand);
			widgets.line_block_on_pitch_rand_seed = build_gtk_spin_button_new (0, SHRT_MAX, 1, 0, vals->line_block_on_pitch_rand_seed);
			widgets.line_block_off_pitch      = build_gtk_spin_button_new (0, SHRT_MAX, 1, 3, vals->line_block_off_pitch);
			widgets.line_block_off_pitch_rand = build_gtk_spin_button_new (0, 100 , 1, 2, vals->line_block_off_pitch_rand);
			widgets.line_block_off_pitch_rand_seed = build_gtk_spin_button_new (0, SHRT_MAX, 1, 0, vals->line_block_off_pitch_rand_seed);
			widgets.line_wave_npoint          = build_gtk_spin_button_new (0, 128, 4, 0, vals->line_wave_npoint);
			widgets.line_wave_depth           = build_gtk_spin_button_new (0, SHRT_MAX, 1, 3, vals->line_wave_depth);
			widgets.line_wave_depth_rand      = build_gtk_spin_button_new (0, 100, 1, 2, vals->line_wave_depth_rand);
			widgets.line_wave_depth_rand_seed = build_gtk_spin_button_new (0, SHRT_MAX, 1, 0, vals->line_wave_depth_rand_seed);
			widgets.line_wave_type            = build_line_wave_type_combo_box_new (vals->line_wave_type);
			widgets.line_width                = build_gtk_spin_button_new (0.5, SHRT_MAX, 1, 3, vals->line_width);
			widgets.line_width_rand           = build_gtk_spin_button_new (0, 100     , 1, 2, vals->line_width_rand);
			widgets.line_width_rand_seed      = build_gtk_spin_button_new (0, SHRT_MAX, 1, 0, vals->line_width_rand_seed);
			widgets.line_in_margin            = build_gtk_spin_button_new (0, SHRT_MAX, 1, 3, vals->line_in_margin);
			widgets.line_in_margin_rand       = build_gtk_spin_button_new (0, 100     , 1, 2, vals->line_in_margin_rand);
			widgets.line_in_margin_rand_seed  = build_gtk_spin_button_new (0, SHRT_MAX, 1, 0, vals->line_in_margin_rand_seed);
			widgets.line_in_length            = build_gtk_spin_button_new (0, SHRT_MAX, 1, 3, vals->line_in_length);
			widgets.line_in_length_rand       = build_gtk_spin_button_new (0, 100     , 1, 2, vals->line_in_length_rand);
			widgets.line_in_length_rand_seed  = build_gtk_spin_button_new (0, SHRT_MAX, 1, 0, vals->line_in_length_rand_seed);
			widgets.line_out_margin           = build_gtk_spin_button_new (0, SHRT_MAX, 1, 3, vals->line_out_margin);
			widgets.line_out_margin_rand      = build_gtk_spin_button_new (0, 100     , 1, 2, vals->line_out_margin_rand);
			widgets.line_out_margin_rand_seed = build_gtk_spin_button_new (0, SHRT_MAX, 1, 0, vals->line_out_margin_rand_seed);
			widgets.line_out_length           = build_gtk_spin_button_new (0, SHRT_MAX, 1, 3, vals->line_out_length);
			widgets.line_out_length_rand      = build_gtk_spin_button_new (0, 100     , 1, 2, vals->line_out_length_rand);
			widgets.line_out_length_rand_seed = build_gtk_spin_button_new (0, SHRT_MAX, 1, 0, vals->line_out_length_rand_seed);
			
			g_signal_connect (widgets.line_color, "color-changed", G_CALLBACK(line_color_color_changed_callback), vals);
			g_signal_connect (widgets.line_antialias, "changed", G_CALLBACK (line_antialias_changed_callback), vals);
			g_signal_connect (widgets.line_pitch, "value_changed", G_CALLBACK (line_pitch_value_changed_callback), vals);
			g_signal_connect (widgets.line_pitch_rand, "value_changed", G_CALLBACK (line_pitch_rand_value_changed_callback), vals);
			g_signal_connect (widgets.line_pitch_rand_seed, "value_changed", G_CALLBACK (line_pitch_rand_seed_value_changed_callback), vals);
			g_signal_connect (widgets.line_pitch_type, "changed", G_CALLBACK (line_pitch_type_changed_callback), vals);
			g_signal_connect (widgets.line_block_on_pitch, "value_changed", G_CALLBACK (line_block_on_pitch_value_changed_callback), vals);
			g_signal_connect (widgets.line_block_on_pitch_rand, "value_changed", G_CALLBACK (line_block_on_pitch_rand_value_changed_callback), vals);
			g_signal_connect (widgets.line_block_on_pitch_rand_seed, "value_changed", G_CALLBACK (line_block_on_pitch_rand_seed_value_changed_callback), vals);
			g_signal_connect (widgets.line_block_off_pitch, "value_changed", G_CALLBACK (line_block_off_pitch_value_changed_callback), vals);
			g_signal_connect (widgets.line_block_off_pitch_rand, "value_changed", G_CALLBACK (line_block_off_pitch_rand_value_changed_callback), vals);
			g_signal_connect (widgets.line_block_off_pitch_rand_seed, "value_changed", G_CALLBACK (line_block_off_pitch_rand_seed_value_changed_callback), vals);
			g_signal_connect (widgets.line_wave_npoint, "value_changed", G_CALLBACK (line_wave_npoint_value_changed_callback), vals);
			g_signal_connect (widgets.line_wave_depth, "value_changed", G_CALLBACK (line_wave_depth_value_changed_callback), vals);
			g_signal_connect (widgets.line_wave_depth_rand, "value_changed", G_CALLBACK (line_wave_depth_rand_value_changed_callback), vals);
			g_signal_connect (widgets.line_wave_depth_rand_seed, "value_changed", G_CALLBACK (line_wave_depth_rand_seed_value_changed_callback), vals);
			g_signal_connect (widgets.line_wave_type, "changed", G_CALLBACK (line_wave_type_changed_callback), vals);
			g_signal_connect (widgets.line_width, "value_changed", G_CALLBACK (line_width_value_changed_callback), vals);
			g_signal_connect (widgets.line_width_rand, "value_changed", G_CALLBACK (line_width_rand_value_changed_callback), vals);
			g_signal_connect (widgets.line_width_rand_seed, "value_changed", G_CALLBACK (line_width_rand_seed_value_changed_callback), vals);
			g_signal_connect (widgets.line_in_margin, "value_changed", G_CALLBACK (line_in_margin_value_changed_callback), vals);
			g_signal_connect (widgets.line_in_margin_rand, "value_changed", G_CALLBACK (line_in_margin_rand_value_changed_callback), vals);
			g_signal_connect (widgets.line_in_margin_rand_seed, "value_changed", G_CALLBACK (line_in_margin_rand_seed_value_changed_callback), vals);
			g_signal_connect (widgets.line_in_length, "value_changed", G_CALLBACK (line_in_length_value_changed_callback), vals);
			g_signal_connect (widgets.line_in_length_rand, "value_changed", G_CALLBACK (line_in_length_rand_value_changed_callback), vals);
			g_signal_connect (widgets.line_in_length_rand_seed, "value_changed", G_CALLBACK (line_in_length_rand_seed_value_changed_callback), vals);
			g_signal_connect (widgets.line_out_margin, "value_changed", G_CALLBACK (line_out_margin_value_changed_callback), vals);
			g_signal_connect (widgets.line_out_margin_rand, "value_changed", G_CALLBACK (line_out_margin_rand_value_changed_callback), vals);
			g_signal_connect (widgets.line_out_margin_rand_seed, "value_changed", G_CALLBACK (line_out_margin_rand_seed_value_changed_callback), vals);
			g_signal_connect (widgets.line_out_length, "value_changed", G_CALLBACK (line_out_length_value_changed_callback), vals);
			g_signal_connect (widgets.line_out_length_rand, "value_changed", G_CALLBACK (line_out_length_rand_value_changed_callback), vals);
			g_signal_connect (widgets.line_out_length_rand_seed, "value_changed", G_CALLBACK (line_out_length_rand_seed_value_changed_callback), vals);
			
			gtk_table_attach (GTK_TABLE (line_option_table), label_antialias,        0, 1, 0, 1, GTK_FILL, GTK_FILL, PADDING, PADDING);
			gtk_table_attach (GTK_TABLE (line_option_table), widgets.line_antialias, 1, 2, 0, 1, GTK_FILL, GTK_FILL, PADDING, PADDING);
			gtk_table_attach (GTK_TABLE (line_option_table), label_color,            2, 3, 0, 1, GTK_FILL, GTK_FILL, PADDING, PADDING);
			gtk_table_attach (GTK_TABLE (line_option_table), widgets.line_color,     3, 4, 0, 1, GTK_FILL, GTK_FILL, PADDING, PADDING);

			gtk_table_attach (GTK_TABLE (line_option_table), label_amount   , 1, 2, 1, 2, GTK_FILL, GTK_FILL, PADDING, PADDING);
			gtk_table_attach (GTK_TABLE (line_option_table), label_rand     , 2, 3, 1, 2, GTK_FILL, GTK_FILL, PADDING, PADDING);
			gtk_table_attach (GTK_TABLE (line_option_table), label_rand_seed, 3, 4, 1, 2, GTK_FILL, GTK_FILL, PADDING, PADDING);
			gtk_table_attach (GTK_TABLE (line_option_table), label_type     , 4, 5, 1, 2, GTK_FILL, GTK_FILL, PADDING, PADDING);

			gtk_table_attach (GTK_TABLE (line_option_table), label_pitch                 , 0, 1, 2, 3, GTK_FILL, GTK_FILL, PADDING, PADDING);
			gtk_table_attach (GTK_TABLE (line_option_table), widgets.line_pitch          , 1, 2, 2, 3, GTK_FILL, GTK_FILL, PADDING, PADDING);
			gtk_table_attach (GTK_TABLE (line_option_table), widgets.line_pitch_rand     , 2, 3, 2, 3, GTK_FILL, GTK_FILL, PADDING, PADDING);
			gtk_table_attach (GTK_TABLE (line_option_table), widgets.line_pitch_rand_seed, 3, 4, 2, 3, GTK_FILL, GTK_FILL, PADDING, PADDING);
			gtk_table_attach (GTK_TABLE (line_option_table), widgets.line_pitch_type     , 4, 5, 2, 3, GTK_FILL, GTK_FILL, PADDING, PADDING);

			gtk_table_attach (GTK_TABLE (line_option_table), label_block_on_pitch                 , 0, 1, 3, 4, GTK_FILL, GTK_FILL, PADDING, PADDING);
			gtk_table_attach (GTK_TABLE (line_option_table), widgets.line_block_on_pitch          , 1, 2, 3, 4, GTK_FILL, GTK_FILL, PADDING, PADDING);
			gtk_table_attach (GTK_TABLE (line_option_table), widgets.line_block_on_pitch_rand     , 2, 3, 3, 4, GTK_FILL, GTK_FILL, PADDING, PADDING);
			gtk_table_attach (GTK_TABLE (line_option_table), widgets.line_block_on_pitch_rand_seed, 3, 4, 3, 4, GTK_FILL, GTK_FILL, PADDING, PADDING);

			gtk_table_attach (GTK_TABLE (line_option_table), label_block_off_pitch                 , 0, 1, 4, 5, GTK_FILL, GTK_FILL, PADDING, PADDING);
			gtk_table_attach (GTK_TABLE (line_option_table), widgets.line_block_off_pitch          , 1, 2, 4, 5, GTK_FILL, GTK_FILL, PADDING, PADDING);
			gtk_table_attach (GTK_TABLE (line_option_table), widgets.line_block_off_pitch_rand     , 2, 3, 4, 5, GTK_FILL, GTK_FILL, PADDING, PADDING);
			gtk_table_attach (GTK_TABLE (line_option_table), widgets.line_block_off_pitch_rand_seed, 3, 4, 4, 5, GTK_FILL, GTK_FILL, PADDING, PADDING);

			gtk_table_attach (GTK_TABLE (line_option_table), label_wave_npoint                 , 0, 1, 5, 6, GTK_FILL, GTK_FILL, PADDING, PADDING);
			gtk_table_attach (GTK_TABLE (line_option_table), widgets.line_wave_npoint          , 1, 2, 5, 6, GTK_FILL, GTK_FILL, PADDING, PADDING);

			gtk_table_attach (GTK_TABLE (line_option_table), label_wave_depth                 , 0, 1, 6, 7, GTK_FILL, GTK_FILL, PADDING, PADDING);
			gtk_table_attach (GTK_TABLE (line_option_table), widgets.line_wave_depth          , 1, 2, 6, 7, GTK_FILL, GTK_FILL, PADDING, PADDING);
			gtk_table_attach (GTK_TABLE (line_option_table), widgets.line_wave_depth_rand     , 2, 3, 6, 7, GTK_FILL, GTK_FILL, PADDING, PADDING);
			gtk_table_attach (GTK_TABLE (line_option_table), widgets.line_wave_depth_rand_seed, 3, 4, 6, 7, GTK_FILL, GTK_FILL, PADDING, PADDING);
			gtk_table_attach (GTK_TABLE (line_option_table), widgets.line_wave_type           , 4, 5, 6, 7, GTK_FILL, GTK_FILL, PADDING, PADDING);

			gtk_table_attach (GTK_TABLE (line_option_table), label_width                 , 0, 1, 7, 8, GTK_FILL, GTK_FILL, PADDING, PADDING);
			gtk_table_attach (GTK_TABLE (line_option_table), widgets.line_width          , 1, 2, 7, 8, GTK_FILL, GTK_FILL, PADDING, PADDING);
			gtk_table_attach (GTK_TABLE (line_option_table), widgets.line_width_rand     , 2, 3, 7, 8, GTK_FILL, GTK_FILL, PADDING, PADDING);
			gtk_table_attach (GTK_TABLE (line_option_table), widgets.line_width_rand_seed, 3, 4, 7, 8, GTK_FILL, GTK_FILL, PADDING, PADDING);

			gtk_table_attach (GTK_TABLE (line_option_table), label_in_margin                 , 0, 1, 8, 9, GTK_FILL, GTK_FILL, PADDING, PADDING);
			gtk_table_attach (GTK_TABLE (line_option_table), widgets.line_in_margin          , 1, 2, 8, 9, GTK_FILL, GTK_FILL, PADDING, PADDING);
			gtk_table_attach (GTK_TABLE (line_option_table), widgets.line_in_margin_rand     , 2, 3, 8, 9, GTK_FILL, GTK_FILL, PADDING, PADDING);
			gtk_table_attach (GTK_TABLE (line_option_table), widgets.line_in_margin_rand_seed, 3, 4, 8, 9, GTK_FILL, GTK_FILL, PADDING, PADDING);

			gtk_table_attach (GTK_TABLE (line_option_table), label_in_length                 , 0, 1, 9, 10, GTK_FILL, GTK_FILL, PADDING, PADDING);
			gtk_table_attach (GTK_TABLE (line_option_table), widgets.line_in_length          , 1, 2, 9, 10, GTK_FILL, GTK_FILL, PADDING, PADDING);
			gtk_table_attach (GTK_TABLE (line_option_table), widgets.line_in_length_rand     , 2, 3, 9, 10, GTK_FILL, GTK_FILL, PADDING, PADDING);
			gtk_table_attach (GTK_TABLE (line_option_table), widgets.line_in_length_rand_seed, 3, 4, 9, 10, GTK_FILL, GTK_FILL, PADDING, PADDING);

			gtk_table_attach (GTK_TABLE (line_option_table), label_out_margin                 , 0, 1, 10, 11, GTK_FILL, GTK_FILL, PADDING, PADDING);
			gtk_table_attach (GTK_TABLE (line_option_table), widgets.line_out_margin          , 1, 2, 10, 11, GTK_FILL, GTK_FILL, PADDING, PADDING);
			gtk_table_attach (GTK_TABLE (line_option_table), widgets.line_out_margin_rand     , 2, 3, 10, 11, GTK_FILL, GTK_FILL, PADDING, PADDING);
			gtk_table_attach (GTK_TABLE (line_option_table), widgets.line_out_margin_rand_seed, 3, 4, 10, 11, GTK_FILL, GTK_FILL, PADDING, PADDING);

			gtk_table_attach (GTK_TABLE (line_option_table), label_out_length                 , 0, 1, 11, 12, GTK_FILL, GTK_FILL, PADDING, PADDING);
			gtk_table_attach (GTK_TABLE (line_option_table), widgets.line_out_length          , 1, 2, 11, 12, GTK_FILL, GTK_FILL, PADDING, PADDING);
			gtk_table_attach (GTK_TABLE (line_option_table), widgets.line_out_length_rand     , 2, 3, 11, 12, GTK_FILL, GTK_FILL, PADDING, PADDING);
			gtk_table_attach (GTK_TABLE (line_option_table), widgets.line_out_length_rand_seed, 3, 4, 11, 12, GTK_FILL, GTK_FILL, PADDING, PADDING);
		}
	}
	
	//preset
	{
		if (preset_dir == NULL) {
			preset_dir = get_preset_dir ();
		}

		widgets.preset_frame = gtk_frame_new (NULL);
		gtk_frame_set_shadow_type ( GTK_FRAME (widgets.preset_frame), GTK_SHADOW_IN);

		widgets.preset_listbox = gtk_scrolled_window_new (NULL, NULL);
		gtk_scrolled_window_set_policy ( GTK_SCROLLED_WINDOW (widgets.preset_listbox), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
		gtk_widget_set_size_request (widgets.preset_listbox, 200, 80);
		gtk_container_add ( GTK_CONTAINER (widgets.preset_frame), widgets.preset_listbox);

		widgets.preset_store = gtk_list_store_new (1, G_TYPE_STRING);
		widgets.preset_tree = gtk_tree_view_new_with_model (GTK_TREE_MODEL (widgets.preset_store));
		gtk_tree_view_set_headers_visible (GTK_TREE_VIEW (widgets.preset_tree), FALSE);
		gtk_tree_view_append_column (
			GTK_TREE_VIEW (widgets.preset_tree),
			gtk_tree_view_column_new_with_attributes (NULL, gtk_cell_renderer_text_new (), "text", 0, NULL));

		g_signal_connect (widgets.preset_tree, "row-activated", G_CALLBACK (preset_tree_dbclicked), NULL);

		gtk_container_add ( GTK_CONTAINER (widgets.preset_listbox), widgets.preset_tree);

		widgets.preset_load   = gtk_button_new_with_label (_("Load"));
		widgets.preset_save   = gtk_button_new_with_label (_("Save"));
		widgets.preset_delete = gtk_button_new_with_label (_("Delete"));
		widgets.preset_update = gtk_button_new_with_label (_("Update List"));
		g_signal_connect (widgets.preset_load, "clicked", G_CALLBACK (preset_load_callback), NULL);
		g_signal_connect (widgets.preset_save, "clicked", G_CALLBACK (preset_save_callback), vals);
		g_signal_connect (widgets.preset_delete, "clicked", G_CALLBACK (preset_delete_callback), NULL);
		g_signal_connect (widgets.preset_update, "clicked", G_CALLBACK (preset_update_callback), NULL);

		preset_update_callback (widgets.preset_update, NULL);

		gtk_table_attach (GTK_TABLE (preset_table), widgets.preset_frame,  0, 1, 0, 20, GTK_EXPAND|GTK_FILL, GTK_EXPAND|GTK_FILL, 2, 2);
		gtk_table_attach (GTK_TABLE (preset_table), widgets.preset_load,   1, 2, 0, 1, GTK_FILL, 0, 2, 2);
		gtk_table_attach (GTK_TABLE (preset_table), widgets.preset_save,   1, 2, 1, 2, GTK_FILL, 0, 2, 2);
		gtk_table_attach (GTK_TABLE (preset_table), widgets.preset_delete, 1, 2, 2, 3, GTK_FILL, 0, 2, 2);
		gtk_table_attach (GTK_TABLE (preset_table), widgets.preset_update, 1, 2, 3, 4, GTK_FILL, 0, 2, 2);
	}

	//
	// show all dialog
	//
	if (vals->gui.width > 100 && vals->gui.height > 100) {
		gtk_window_resize (GTK_WINDOW (widgets.dialog), vals->gui.width, vals->gui.height);
	}
	if (vals->gui.x < 0) {
		vals->gui.x = 0;
	}
	if (vals->gui.y < 0) {
		vals->gui.y = 0;
	}
	if (vals->gui.x == 0 && vals->gui.y == 0) {
		gtk_window_set_position (GTK_WINDOW (widgets.dialog), GTK_WIN_POS_CENTER_ALWAYS);
	} else {
		gtk_window_move (GTK_WINDOW (widgets.dialog), vals->gui.x, vals->gui.y);
	}

	gtk_widget_show_all (widgets.dialog);
	//gtk_note_book_set_current_page () requires his child widgets are shown.
	gtk_notebook_set_current_page (GTK_NOTEBOOK (widgets.area_type_notebook), vals->area_type);
	gtk_window_set_keep_above (GTK_WINDOW (widgets.dialog), vals->gui.topmost);

	//hide_splash ();
	
	gtk_main ();
	return ERROR_NOTHING;
#undef PADDING
}/*}}}*/

static int
do_speedline (PluginVals	*vals,
			  gint32		*out_layer_ID)
{/*{{{*/
	int			selection_ID = 0;
	gboolean	have_selection;
	gint		width, height;
	/* Layer */
	gint32			layer_ID = 0;
	GimpDrawable	*drawable = NULL;
	char			*layer_name = NULL;
	GimpPixelRgn	rgn;
	guint8			*buf = NULL;
	GimpImageType	layer_type;

	cairo_surface_t	*surface = NULL;
	cairo_t			*cr = NULL;
	guint8			*surface_data =NULL;

	int error = ERROR_NOTHING;

	have_selection = !gimp_selection_is_empty (vals->image_ID);
	if (have_selection) {
		gboolean non_empty;
		gint sx1, sy1, sx2, sy2;
		gimp_selection_bounds (vals->image_ID, &non_empty, &sx1, &sy1, &sx2, &sy2);
		width = sx2 - sx1;
		height = sy2 - sy1;
	} else {
		width = gimp_image_width (vals->image_ID);
		height = gimp_image_height (vals->image_ID);
	}

	surface = cairo_image_surface_create (CAIRO_FORMAT_ARGB32, width, height);

	cr = cairo_create (surface);
	switch (vals->area_type) {
		case AREA_TYPE_RADIATION: radiation_draw_line (cr, vals); break;
		case AREA_TYPE_PARALLEL:  parallel_draw_line (cr, vals); break;
	}
	cairo_surface_flush (surface);
	surface_data = cairo_image_surface_get_data (surface); //dont free surface_data. It is owned by image surface.
	{
		int s_stride = cairo_image_surface_get_stride (surface);
		int s_height = cairo_image_surface_get_height (surface);
		int r, c;
		guint8 rgb[3];
		guint8 *dst;

		//ignore cairo's renderging color. use alpha plane only.
		// because of juggy antialias edge.

		switch (gimp_image_base_type (vals->image_ID)) {
			case GIMP_RGB:
				gimp_rgb_get_uchar (&vals->line_color, rgb, rgb+1, rgb+2);
				buf = g_new0 (guint8, width * height * 4);
				dst = buf;
				for (r = 0; r < height * s_stride; r += s_stride) {
					for (c = 0; c < width * 4; c+= 4) {
						*dst++ = rgb[0];
						*dst++ = rgb[1];
						*dst++ = rgb[2];
						*dst++ = surface_data[r + c + 3];//a
					}
				}
				layer_type =GIMP_RGBA_IMAGE;
				break;
			case GIMP_GRAY:
				rgb[0] = gimp_rgb_intensity_uchar (&vals->line_color);
				buf = g_new0 (guint8, width * height * 2);
				dst = buf;
				for (r = 0; r < height * s_stride; r += s_stride) {
					for (c = 0; c < width * 4; c+= 4) {
						*dst++ = rgb[0];
						*dst++ = surface_data[r + c + 3];
					}
				}
				layer_type =GIMP_GRAYA_IMAGE;
				break;
			default:
				error = ERROR_UNKNOWN_IMAGE_TYPE_ERROR;
				goto done;
		}
	}

	cairo_destroy (cr);
	cairo_surface_destroy (surface);
	cr = NULL;
	surface = NULL;


	gimp_image_undo_group_start (vals->image_ID);

	//add layer
	if (vals->mode == RUNMODE_ADD) {
		// create new layer and copy the image
		layer_name = create_layer_name ("Speedline Layer");
		layer_ID = gimp_layer_new (
				vals->image_ID,
				layer_name,
				width,
				height,
				layer_type,
				100,
				GIMP_NORMAL_MODE);
		if (layer_name) {
			g_free (layer_name);
			layer_name = NULL;
		}
		if (layer_ID < 0) {
			WNGMSG ("Error:Failed to create new layer");
			error = ERROR_FAILED_CREATE_LAYER_ERROR;
			goto done;
		}
		{
			guint32 active_layer;
			active_layer = gimp_image_get_active_layer (vals->image_ID);
			if (gimp_item_is_group (active_layer)) {
				gimp_image_insert_layer (vals->image_ID, layer_ID, active_layer, -1);
			} else {
				guint32 parent_ID = gimp_item_get_parent (active_layer);
				int pos = gimp_image_get_layer_position (vals->image_ID, active_layer);
				gimp_image_insert_layer (vals->image_ID, layer_ID, parent_ID, pos);
			}
			gimp_image_set_active_layer (vals->image_ID, layer_ID);
		}

		//offset layer to selection rectangle
		if (!have_selection) {
			gimp_layer_set_offsets (layer_ID, 0, 0);
		} else {
			gboolean non_empty;
			gint sx1, sy1, sx2, sy2;
			gimp_selection_bounds (vals->image_ID, &non_empty, &sx1, &sy1, &sx2, &sy2);
			gimp_layer_set_offsets (layer_ID, sx2 - width, sy1);

			selection_ID = gimp_selection_save (vals->image_ID);
			gimp_selection_none (vals->image_ID);
		}
	}

	//modify layer
	if (vals->mode == RUNMODE_MODIFY) {
		int oldwidth, oldheight;
		layer_ID = vals->layer_ID;

		layer_name = create_layer_name ("Speedline Layer");
		gimp_item_set_name (layer_ID, layer_name);

		if (layer_name) {
			g_free (layer_name);
			layer_name = NULL;
		}

		oldwidth = gimp_drawable_width (layer_ID);
		oldheight = gimp_drawable_height (layer_ID);

		gimp_layer_resize (layer_ID, width, height, width - oldwidth, 0);
	}
	
	gimp_layer_add_alpha (layer_ID);
	drawable = gimp_drawable_get (layer_ID);
	
	//gimp_tile_cache_ntiles ( 4 * (drawable->width / gimp_tile_width () + 1));

	gimp_pixel_rgn_init (&rgn, drawable, 0, 0, width, height, TRUE, TRUE);

	gimp_pixel_rgn_set_rect (&rgn, buf, 0, 0, width, height);

	gimp_drawable_flush (drawable);

	gimp_drawable_merge_shadow ( drawable->drawable_id, TRUE);

	gimp_drawable_update (drawable->drawable_id, 0, 0, drawable->width, drawable->height);

	gimp_displays_flush ();

	gimp_drawable_detach (drawable);


	// attatch parasite
	/*
	{
		GimpParasite *parasite;
		parasite = pluginvals_to_parasite (vals);
		if (parasite != NULL) {
			gimp_item_detach_parasite (layer_ID, PLUGIN_ID);
			gimp_item_attach_parasite (layer_ID, parasite);
		}
		gimp_parasite_free (parasite);
	}
	*/

	if (vals->mode == RUNMODE_ADD) {
		if (have_selection) {
			gimp_selection_load (selection_ID);
			gimp_image_remove_channel (vals->image_ID, selection_ID);
		}
	}

	gimp_image_undo_group_end (vals->image_ID);


done:
	if (layer_name) g_free (layer_name);
	*out_layer_ID = (error == ERROR_NOTHING) ? layer_ID : -1;
	return error;
}/*}}}*/

static char*
create_layer_name (char *text)
{/*{{{*/
	gboolean overflow = 0;
	int all_num, cpy_num;
	char *dst, *ret;
	all_num = g_utf8_strlen (text, -1);
	overflow = all_num > 20;
	cpy_num = overflow ? 20 : all_num;
	
	dst = ret = g_malloc0 (64);
	g_utf8_strncpy (dst, text, cpy_num);
	dst += strlen (dst);
	if (overflow) {
		g_utf8_strncpy (dst, "...", 3);
	}
	
	dst = ret;
	while (*dst) {
		if (*dst == 0x0a) {
			*dst = 0x2F;
		}
		dst = g_utf8_next_char (dst);
	}
	return ret;
}/*}}}*/

static void
reset_default_vals (PluginVals *vals)
{/*{{{*/
	gboolean nonempty;
	gint rx1, ry1, rx2, ry2;

	if (gimp_selection_is_empty (vals->image_ID)) {
		rx1 = 0;
		ry1 = 0;
		rx2 = gimp_image_width (vals->image_ID);
		ry2 = gimp_image_height (vals->image_ID);
	} else {
		gimp_selection_bounds (vals->image_ID, &nonempty, &rx1, &ry1, &rx2, &ry2);
	}

	vals->area_type = AREA_TYPE_RADIATION;
	vals->area_unit = GIMP_UNIT_PIXEL;

	vals->radiation_x = (rx2 - rx1) / 2;
	vals->radiation_y = (ry2 - ry1) / 2;
	vals->radiation_out_radius = sqrt (pow (rx2 - rx1, 2) + pow (ry2 - ry1, 2)) / 2;
	vals->radiation_in_radius = vals->radiation_out_radius / 2;
	vals->radiation_oblateness = 0;
	vals->radiation_in_freehand_active = 0;
	radiation_reset_freehand (vals->radiation_in_freehand_points,
							 &vals->radiation_in_freehand_points_count);
	vals->radiation_out_freehand_active = 0;
	radiation_reset_freehand (vals->radiation_out_freehand_points,
							 &vals->radiation_out_freehand_points_count);

	vals->parallel_in_x = 0;
	vals->parallel_in_y = (ry2 - ry1) / 2;
	vals->parallel_in_range = ry2 - ry1;
	vals->parallel_out_x = rx2 - rx1;
	vals->parallel_out_y = (ry2 - ry1) / 2;
	vals->parallel_out_range = ry2 - ry1;
	vals->parallel_in_freehand_active = 0;
	parallel_reset_freehand (vals->parallel_in_freehand_points,
							&vals->parallel_in_freehand_points_count);
	vals->parallel_out_freehand_active = 0;
	parallel_reset_freehand (vals->parallel_out_freehand_points,
							&vals->parallel_out_freehand_points_count);

	vals->line_color.r = 0;
	vals->line_color.g = 0;
	vals->line_color.b = 0;
	vals->line_color.a = 1;
	vals->line_antialias = ANTIALIAS_TYPE_BEST;
	vals->line_pitch = 10;
	vals->line_pitch_rand = 100;
	vals->line_pitch_rand_seed = 100;
	vals->line_pitch_type = LINE_PITCH_TYPE_OUT;
	vals->line_block_on_pitch = 0;
	vals->line_block_on_pitch_rand = 0;
	vals->line_block_on_pitch_rand_seed = 100;
	vals->line_block_off_pitch = 0;
	vals->line_block_off_pitch_rand = 0;
	vals->line_block_off_pitch_rand_seed = 100;
	vals->line_wave_npoint = 0;
	vals->line_wave_depth = 50;
	vals->line_wave_depth_rand = 0;
	vals->line_wave_depth_rand_seed = 100;
	vals->line_wave_type = LINE_WAVE_TYPE_DEFAULT;
	vals->line_width = 2;
	vals->line_width_rand = 0;
	vals->line_width_rand_seed = 100;
	vals->line_in_margin = (vals->radiation_out_radius - vals->radiation_in_radius) / 2;
	vals->line_in_margin_rand = 50;
	vals->line_in_margin_rand_seed = 100;
	vals->line_in_length = (vals->radiation_out_radius - vals->radiation_in_radius) / 2;
	vals->line_in_length_rand = 0;
	vals->line_in_length_rand_seed = 100;
	vals->line_out_margin = 0;
	vals->line_out_margin_rand = 0;
	vals->line_out_margin_rand_seed = 100;
	vals->line_out_length = 0;
	vals->line_out_length_rand = 0;
	vals->line_out_length_rand_seed = 100;
}/*}}}*/

static void
parasite_to_widgets (GimpParasite *parasite)
{/*{{{*/
	gchar* data;
	gint32 size;
	GKeyFile *kfile;
	gboolean result;
	if (parasite == NULL) return;

	data = (gchar*)gimp_parasite_data (parasite);
	size = gimp_parasite_data_size (parasite);

	kfile = g_key_file_new ();
	result = g_key_file_load_from_data (kfile, data, size, G_KEY_FILE_NONE, NULL);
	if (kfile != NULL) {
		keyfile_to_widgets (kfile);
		g_key_file_free (kfile);
	}
}/*}}}*/

static GimpParasite*
pluginvals_to_parasite (PluginVals *vals)
{/*{{{*/
	GKeyFile *kfile;
	gchar *data;
	GimpParasite *parasite;

	kfile = pluginvals_to_keyfile (vals);

	{
		gchar **list;
		list = g_key_file_get_keys (kfile, PNAME_GROUP, NULL, NULL);
	}
	data = g_key_file_to_data (kfile, NULL, NULL);
	parasite = gimp_parasite_new (PLUGIN_ID, 
								GIMP_PARASITE_PERSISTENT | GIMP_PARASITE_UNDOABLE,
								(guint32)strlen (data) * sizeof (gchar), data);

	g_free (data);
	g_key_file_free (kfile);

	return parasite;
}/*}}}*/ 

static GKeyFile*
pluginvals_to_keyfile (PluginVals *vals)
{/*{{{*/
#define SET_FREEHAND_POINTS(KEY, POINTS, COUNT) \
if ((COUNT) > 0) { \
	int i; \
	int len = (COUNT); \
	gdouble *points; \
	points = g_malloc_n (FREEHAND_POINTS_MAX * 2, sizeof(gdouble)); \
	for (i = 0; i < len; i++) { \
		points[i * 2]     = (POINTS)[i].key; \
		points[i * 2 + 1] = (POINTS)[i].value; \
	} \
	g_key_file_set_double_list (kfile, PNAME_GROUP, (KEY), points, len); \
	g_free (points); \
}

	GKeyFile *kfile;
	guchar rgb[3];
	gint   list[3];

	kfile = g_key_file_new ();

	g_key_file_set_string (kfile, PNAME_GROUP, PNAME_VERSION, PLUGIN_VERSION);

	gimp_rgb_get_uchar (&vals->line_color, rgb, rgb+1, rgb+2);
	list[0] = (gint)rgb[0];
	list[1] = (gint)rgb[1];
	list[2] = (gint)rgb[2];
	g_key_file_set_integer_list (kfile, PNAME_GROUP, PNAME_LINE_COLOR, list, 3);
	g_key_file_set_integer (kfile, PNAME_GROUP, PNAME_LINE_ANTIALIAS, vals->line_antialias);
	g_key_file_set_double (kfile, PNAME_GROUP, PNAME_LINE_PITCH, vals->line_pitch);
	g_key_file_set_double (kfile, PNAME_GROUP, PNAME_LINE_PITCH_RAND, vals->line_pitch_rand);
	g_key_file_set_integer (kfile, PNAME_GROUP, PNAME_LINE_PITCH_RAND_SEED, vals->line_pitch_rand_seed);
	g_key_file_set_integer (kfile, PNAME_GROUP, PNAME_LINE_PITCH_TYPE, vals->line_pitch_type);
	g_key_file_set_double (kfile, PNAME_GROUP, PNAME_LINE_BLOCK_ON_PITCH, vals->line_block_on_pitch);
	g_key_file_set_double (kfile, PNAME_GROUP, PNAME_LINE_BLOCK_ON_PITCH_RAND, vals->line_block_on_pitch_rand);
	g_key_file_set_integer (kfile, PNAME_GROUP, PNAME_LINE_BLOCK_ON_PITCH_RAND_SEED, vals->line_block_on_pitch_rand_seed);
	g_key_file_set_double (kfile, PNAME_GROUP, PNAME_LINE_BLOCK_OFF_PITCH, vals->line_block_off_pitch);
	g_key_file_set_double (kfile, PNAME_GROUP, PNAME_LINE_BLOCK_OFF_PITCH_RAND, vals->line_block_off_pitch_rand);
	g_key_file_set_integer (kfile, PNAME_GROUP, PNAME_LINE_BLOCK_OFF_PITCH_RAND_SEED, vals->line_block_off_pitch_rand_seed);
	g_key_file_set_integer (kfile, PNAME_GROUP, PNAME_LINE_WAVE_NPOINT, vals->line_wave_npoint);
	g_key_file_set_double (kfile, PNAME_GROUP, PNAME_LINE_WAVE_DEPTH, vals->line_wave_depth);
	g_key_file_set_double (kfile, PNAME_GROUP, PNAME_LINE_WAVE_DEPTH_RAND, vals->line_wave_depth_rand);
	g_key_file_set_integer (kfile, PNAME_GROUP, PNAME_LINE_WAVE_DEPTH_RAND_SEED, vals->line_wave_depth_rand_seed);
	g_key_file_set_integer (kfile, PNAME_GROUP, PNAME_LINE_WAVE_TYPE, vals->line_wave_type);
	g_key_file_set_double (kfile, PNAME_GROUP, PNAME_LINE_WIDTH, vals->line_width);
	g_key_file_set_double (kfile, PNAME_GROUP, PNAME_LINE_WIDTH_RAND, vals->line_width_rand);
	g_key_file_set_integer (kfile, PNAME_GROUP, PNAME_LINE_WIDTH_RAND_SEED, vals->line_width_rand_seed);
	g_key_file_set_double (kfile, PNAME_GROUP, PNAME_LINE_IN_MARGIN, vals->line_in_margin);
	g_key_file_set_double (kfile, PNAME_GROUP, PNAME_LINE_IN_MARGIN_RAND, vals->line_in_margin_rand);
	g_key_file_set_integer (kfile, PNAME_GROUP, PNAME_LINE_IN_MARGIN_RAND_SEED, vals->line_in_margin_rand_seed);
	g_key_file_set_double (kfile, PNAME_GROUP, PNAME_LINE_IN_LENGTH, vals->line_in_length);
	g_key_file_set_double (kfile, PNAME_GROUP, PNAME_LINE_IN_LENGTH_RAND, vals->line_in_length_rand);
	g_key_file_set_integer (kfile, PNAME_GROUP, PNAME_LINE_IN_LENGTH_RAND_SEED, vals->line_in_length_rand_seed);
	g_key_file_set_double (kfile, PNAME_GROUP, PNAME_LINE_OUT_MARGIN, vals->line_out_margin);
	g_key_file_set_double (kfile, PNAME_GROUP, PNAME_LINE_OUT_MARGIN_RAND, vals->line_out_margin_rand);
	g_key_file_set_integer (kfile, PNAME_GROUP, PNAME_LINE_OUT_MARGIN_RAND_SEED, vals->line_out_margin_rand_seed);
	g_key_file_set_double (kfile, PNAME_GROUP, PNAME_LINE_OUT_LENGTH, vals->line_out_length);
	g_key_file_set_double (kfile, PNAME_GROUP, PNAME_LINE_OUT_LENGTH_RAND, vals->line_out_length_rand);
	g_key_file_set_integer (kfile, PNAME_GROUP, PNAME_LINE_OUT_LENGTH_RAND_SEED, vals->line_out_length_rand_seed);

	g_key_file_set_integer (kfile, PNAME_GROUP, PNAME_AREA_TYPE, vals->area_type);
	g_key_file_set_integer (kfile, PNAME_GROUP, PNAME_AREA_UNIT, vals->area_unit);

	g_key_file_set_double (kfile, PNAME_GROUP, PNAME_RADIATION_X, vals->radiation_x);
	g_key_file_set_double (kfile, PNAME_GROUP, PNAME_RADIATION_Y, vals->radiation_y);
	g_key_file_set_double (kfile, PNAME_GROUP, PNAME_RADIATION_IN_RADIUS, vals->radiation_in_radius);
	g_key_file_set_double (kfile, PNAME_GROUP, PNAME_RADIATION_OUT_RADIUS, vals->radiation_out_radius);
	g_key_file_set_double (kfile, PNAME_GROUP, PNAME_RADIATION_OBLATENESS, vals->radiation_oblateness);
	g_key_file_set_boolean(kfile, PNAME_GROUP, PNAME_RADIATION_IN_FREEHAND_ACTIVE, vals->radiation_in_freehand_active);
	if (vals->radiation_in_freehand_active) {
		g_key_file_set_integer(kfile, PNAME_GROUP, PNAME_RADIATION_IN_FREEHAND_POINTS_COUNT, vals->radiation_in_freehand_points_count);
		SET_FREEHAND_POINTS(PNAME_RADIATION_IN_FREEHAND_POINTS, vals->radiation_in_freehand_points, vals->radiation_in_freehand_points_count);
	}
	g_key_file_set_boolean(kfile, PNAME_GROUP, PNAME_RADIATION_OUT_FREEHAND_ACTIVE, vals->radiation_out_freehand_active);
	if (vals->radiation_out_freehand_active) {
		g_key_file_set_integer(kfile, PNAME_GROUP, PNAME_RADIATION_OUT_FREEHAND_POINTS_COUNT, vals->radiation_out_freehand_points_count);
		SET_FREEHAND_POINTS(PNAME_RADIATION_OUT_FREEHAND_POINTS, vals->radiation_out_freehand_points, vals->radiation_out_freehand_points_count);
	}

	g_key_file_set_double (kfile, PNAME_GROUP, PNAME_PARALLEL_IN_X, vals->parallel_in_x);
	g_key_file_set_double (kfile, PNAME_GROUP, PNAME_PARALLEL_IN_Y, vals->parallel_in_y);
	g_key_file_set_double (kfile, PNAME_GROUP, PNAME_PARALLEL_IN_RANGE, vals->parallel_in_range);
	g_key_file_set_double (kfile, PNAME_GROUP, PNAME_PARALLEL_OUT_X, vals->parallel_out_x);
	g_key_file_set_double (kfile, PNAME_GROUP, PNAME_PARALLEL_OUT_Y, vals->parallel_out_y);
	g_key_file_set_double (kfile, PNAME_GROUP, PNAME_PARALLEL_OUT_RANGE, vals->parallel_out_range);
	g_key_file_set_boolean(kfile, PNAME_GROUP, PNAME_PARALLEL_IN_FREEHAND_ACTIVE, vals->parallel_in_freehand_active);
	if (vals->parallel_in_freehand_active) {
		g_key_file_set_integer(kfile, PNAME_GROUP, PNAME_PARALLEL_IN_FREEHAND_POINTS_COUNT, vals->parallel_in_freehand_points_count);
		SET_FREEHAND_POINTS(PNAME_PARALLEL_IN_FREEHAND_POINTS, vals->parallel_in_freehand_points, vals->parallel_in_freehand_points_count);
	}
	g_key_file_set_boolean(kfile, PNAME_GROUP, PNAME_PARALLEL_OUT_FREEHAND_ACTIVE, vals->parallel_out_freehand_active);
	if (vals->parallel_out_freehand_active) {
		g_key_file_set_integer(kfile, PNAME_GROUP, PNAME_PARALLEL_OUT_FREEHAND_POINTS_COUNT, vals->parallel_out_freehand_points_count);
		SET_FREEHAND_POINTS(PNAME_PARALLEL_OUT_FREEHAND_POINTS, vals->parallel_out_freehand_points, vals->parallel_out_freehand_points_count);
	}

	return kfile;
#undef SET_FREEHAND_POINTS
}/*}}}*/

static PluginVals*
keyfile_to_pluginvals (GKeyFile *kfile)
{/*{{{*/
	PluginVals *vals;
	gint *list;
	GimpRGB color;
	int v_major, v_minor, v_patch;

	vals = g_new0 (PluginVals, 1);

#define GET_BOOL(param, key) \
if (g_key_file_has_key (kfile, PNAME_GROUP, (key), NULL) == TRUE) { \
	(param) = g_key_file_get_boolean (kfile, PNAME_GROUP, (key), NULL); \
}
#define GET_INT(param, key) \
if (g_key_file_has_key (kfile, PNAME_GROUP, (key), NULL) == TRUE) { \
	(param) = g_key_file_get_integer (kfile, PNAME_GROUP, (key), NULL); \
}
#define GET_DOUBLE(param, key) \
if (g_key_file_has_key (kfile, PNAME_GROUP, (key), NULL) == TRUE) { \
	(param) = g_key_file_get_double (kfile, PNAME_GROUP, (key), NULL); \
}
#define GET_VERSION(major, minor, patch, key) \
if (g_key_file_has_key (kfile, PNAME_GROUP, (key), NULL) == TRUE) { \
        char* kfile_version = NULL; \
        char **versions = NULL; \
        kfile_version = g_key_file_get_string (kfile, PNAME_GROUP, (key), NULL); \
        versions = g_strsplit (kfile_version, ".", 3); \
        (major) = atoi(versions[0]); \
        (minor) = atoi(versions[1]); \
        (patch) = atoi(versions[2]); \
        g_strfreev (versions); \
        g_free(kfile_version); \
} else { \
	(major) = 0; \
	(minor) = 0; \
	(patch) = 0; \
}
#define GET_FREEHAND_POINTS(POINTS, COUNT, KEY) \
if (g_key_file_has_key (kfile, PNAME_GROUP, (KEY), NULL) == TRUE) { \
	if ((COUNT) > 0 && (COUNT) == FREEHAND_POINTS_MAX) { \
		gint i; \
		gsize len; \
		gdouble *points; \
		points = g_key_file_get_double_list (kfile, PNAME_GROUP, (KEY), &len, NULL); \
		if (len == FREEHAND_POINTS_MAX * 2) { \
			for (i = 0; i < FREEHAND_POINTS_MAX; i++) { \
				(POINTS)[i].key = points[i * 2]; \
				(POINTS)[i].value = points[i * 2 + 1]; \
			} \
		} else { \
			(COUNT) = 0; \
		} \
	} else { \
		(COUNT) = 0; \
	} \
}

	if (g_key_file_has_key (kfile, PNAME_GROUP, PNAME_LINE_COLOR, NULL) == TRUE) {
		list = g_key_file_get_integer_list (kfile, PNAME_GROUP, PNAME_LINE_COLOR, NULL, NULL);
		gimp_rgb_set_uchar (&color, list[0], list[1], list[2]);
		color.a = 1;
		vals->line_color = color;
		if (list) {
			g_free (list);
			list = NULL;
		}
	}

    GET_VERSION (v_major, v_minor, v_patch, PNAME_VERSION)
    
	GET_INT		(vals->line_antialias,				PNAME_LINE_ANTIALIAS)
	GET_DOUBLE	(vals->line_pitch,					PNAME_LINE_PITCH)
	GET_DOUBLE	(vals->line_pitch_rand,				PNAME_LINE_PITCH_RAND)
	GET_INT		(vals->line_pitch_rand_seed,		PNAME_LINE_PITCH_RAND_SEED)
	GET_INT		(vals->line_pitch_type,				PNAME_LINE_PITCH_TYPE)
	GET_DOUBLE	(vals->line_block_on_pitch,			PNAME_LINE_BLOCK_ON_PITCH);
	GET_DOUBLE	(vals->line_block_on_pitch_rand,	PNAME_LINE_BLOCK_ON_PITCH_RAND);
	GET_INT		(vals->line_block_on_pitch_rand_seed, PNAME_LINE_BLOCK_ON_PITCH_RAND_SEED);
	GET_DOUBLE	(vals->line_block_off_pitch,		PNAME_LINE_BLOCK_OFF_PITCH);
	GET_DOUBLE	(vals->line_block_off_pitch_rand,	PNAME_LINE_BLOCK_OFF_PITCH_RAND);
	GET_INT		(vals->line_block_off_pitch_rand_seed, PNAME_LINE_BLOCK_OFF_PITCH_RAND_SEED);
	GET_INT		(vals->line_wave_npoint,			PNAME_LINE_WAVE_NPOINT);
	GET_DOUBLE	(vals->line_wave_depth,				PNAME_LINE_WAVE_DEPTH);
	GET_DOUBLE	(vals->line_wave_depth_rand,		PNAME_LINE_WAVE_DEPTH_RAND);
	GET_INT		(vals->line_wave_depth_rand_seed,	PNAME_LINE_WAVE_DEPTH_RAND_SEED);
	GET_INT		(vals->line_wave_type,				PNAME_LINE_WAVE_TYPE);
	GET_DOUBLE	(vals->line_width,					PNAME_LINE_WIDTH)
	GET_DOUBLE	(vals->line_width_rand,				PNAME_LINE_WIDTH_RAND)
	GET_INT		(vals->line_width_rand_seed,		PNAME_LINE_WIDTH_RAND_SEED)
	GET_DOUBLE	(vals->line_in_margin,				PNAME_LINE_IN_MARGIN)
	GET_DOUBLE	(vals->line_in_margin_rand,			PNAME_LINE_IN_MARGIN_RAND)
	GET_INT		(vals->line_in_margin_rand_seed,	PNAME_LINE_IN_MARGIN_RAND_SEED)
	GET_DOUBLE	(vals->line_in_length,				PNAME_LINE_IN_LENGTH)
	GET_DOUBLE	(vals->line_in_length_rand,			PNAME_LINE_IN_LENGTH_RAND)
	GET_INT		(vals->line_in_length_rand_seed,	PNAME_LINE_IN_LENGTH_RAND_SEED)
	GET_DOUBLE	(vals->line_out_margin,				PNAME_LINE_OUT_MARGIN)
	GET_DOUBLE	(vals->line_out_margin_rand,		PNAME_LINE_OUT_MARGIN_RAND)
	GET_INT		(vals->line_out_margin_rand_seed,	PNAME_LINE_OUT_MARGIN_RAND_SEED)
	GET_DOUBLE	(vals->line_out_length,				PNAME_LINE_OUT_LENGTH)
	GET_DOUBLE	(vals->line_out_length_rand,		PNAME_LINE_OUT_LENGTH_RAND)
	GET_INT		(vals->line_out_length_rand_seed,	PNAME_LINE_OUT_LENGTH_RAND_SEED)

	GET_INT   (vals->area_type,					PNAME_AREA_TYPE)
	GET_INT   (vals->area_unit,					PNAME_AREA_UNIT)

	GET_DOUBLE(vals->radiation_x,				PNAME_RADIATION_X)
	GET_DOUBLE(vals->radiation_y,				PNAME_RADIATION_Y)

	if (v_major <= 0 && v_minor <= 2) { 
		GET_DOUBLE(vals->radiation_in_radius,		PNAME_RADIATION_OUT_RADIUS)
		GET_DOUBLE(vals->radiation_out_radius,		PNAME_RADIATION_IN_RADIUS)
	} else {
		GET_DOUBLE(vals->radiation_in_radius,		PNAME_RADIATION_IN_RADIUS)
		GET_DOUBLE(vals->radiation_out_radius,		PNAME_RADIATION_OUT_RADIUS)
		GET_DOUBLE(vals->radiation_oblateness,		PNAME_RADIATION_OBLATENESS)
	}
	GET_BOOL(vals->radiation_in_freehand_active, PNAME_RADIATION_IN_FREEHAND_ACTIVE);
	GET_INT(vals->radiation_in_freehand_points_count, PNAME_RADIATION_IN_FREEHAND_POINTS_COUNT);
	GET_FREEHAND_POINTS(vals->radiation_in_freehand_points, vals->radiation_in_freehand_points_count, PNAME_RADIATION_IN_FREEHAND_POINTS);
	GET_BOOL(vals->radiation_out_freehand_active, PNAME_RADIATION_OUT_FREEHAND_ACTIVE);
	GET_INT(vals->radiation_out_freehand_points_count, PNAME_RADIATION_OUT_FREEHAND_POINTS_COUNT);
	GET_FREEHAND_POINTS (vals->radiation_out_freehand_points, vals->radiation_out_freehand_points_count, PNAME_RADIATION_OUT_FREEHAND_POINTS);

	GET_DOUBLE(vals->parallel_in_x,				PNAME_PARALLEL_IN_X)
	GET_DOUBLE(vals->parallel_in_y,				PNAME_PARALLEL_IN_Y)
	GET_DOUBLE(vals->parallel_in_range,			PNAME_PARALLEL_IN_RANGE)
	GET_DOUBLE(vals->parallel_out_x,			PNAME_PARALLEL_OUT_X)
	GET_DOUBLE(vals->parallel_out_y,			PNAME_PARALLEL_OUT_Y)
	GET_DOUBLE(vals->parallel_out_range,		PNAME_PARALLEL_OUT_RANGE)
	GET_BOOL(vals->parallel_in_freehand_active, PNAME_PARALLEL_IN_FREEHAND_ACTIVE);
	GET_INT(vals->parallel_in_freehand_points_count, PNAME_PARALLEL_IN_FREEHAND_POINTS_COUNT);
	GET_FREEHAND_POINTS(vals->parallel_in_freehand_points, vals->parallel_in_freehand_points_count, PNAME_PARALLEL_IN_FREEHAND_POINTS);
	GET_BOOL(vals->parallel_out_freehand_active, PNAME_PARALLEL_OUT_FREEHAND_ACTIVE);
	GET_INT(vals->parallel_out_freehand_points_count, PNAME_PARALLEL_OUT_FREEHAND_POINTS_COUNT);
	GET_FREEHAND_POINTS (vals->parallel_out_freehand_points, vals->parallel_out_freehand_points_count, PNAME_PARALLEL_OUT_FREEHAND_POINTS);

	return vals;
#undef GET_BOOL
#undef GET_INT
#undef GET_DOUBLE
#undef GET_FREEHAND_POINTS
}/*}}}*/

static void
pluginvals_to_widgets (PluginVals *vals)
{/*{{{*/
#define SET_GIMP_INT_COMBO_BOX(VAL, WIDGET) \
gimp_int_combo_box_set_active (GIMP_INT_COMBO_BOX ((WIDGET)), (VAL)); 

#define SET_GTK_SPIN_BUTTON(VAL, WIDGET) \
gtk_spin_button_set_value (GTK_SPIN_BUTTON ((WIDGET)), (VAL));

	gimp_color_button_set_color (GIMP_COLOR_BUTTON (widgets.line_color), &vals->line_color);
	SET_GIMP_INT_COMBO_BOX	(vals->line_antialias,					widgets.line_antialias);
	SET_GTK_SPIN_BUTTON		(vals->line_pitch,						widgets.line_pitch);
	SET_GTK_SPIN_BUTTON		(vals->line_pitch_rand,					widgets.line_pitch_rand);
	SET_GTK_SPIN_BUTTON		(vals->line_pitch_rand_seed,			widgets.line_pitch_rand_seed);
	SET_GIMP_INT_COMBO_BOX	(vals->line_pitch_type,					widgets.line_pitch_type);
	SET_GTK_SPIN_BUTTON		(vals->line_block_on_pitch,				widgets.line_block_on_pitch);
	SET_GTK_SPIN_BUTTON		(vals->line_block_on_pitch_rand,		widgets.line_block_on_pitch_rand);
	SET_GTK_SPIN_BUTTON		(vals->line_block_on_pitch_rand_seed,	widgets.line_block_on_pitch_rand_seed);
	SET_GTK_SPIN_BUTTON		(vals->line_block_off_pitch,			widgets.line_block_off_pitch);
	SET_GTK_SPIN_BUTTON		(vals->line_block_off_pitch_rand,		widgets.line_block_off_pitch_rand);
	SET_GTK_SPIN_BUTTON		(vals->line_block_off_pitch_rand_seed,	widgets.line_block_off_pitch_rand_seed);
	SET_GTK_SPIN_BUTTON		(vals->line_wave_npoint,				widgets.line_wave_npoint);
	SET_GTK_SPIN_BUTTON		(vals->line_wave_depth,					widgets.line_wave_depth);
	SET_GTK_SPIN_BUTTON		(vals->line_wave_depth_rand,			widgets.line_wave_depth_rand);
	SET_GTK_SPIN_BUTTON		(vals->line_wave_depth_rand_seed,		widgets.line_wave_depth_rand_seed);
	SET_GIMP_INT_COMBO_BOX	(vals->line_wave_type,					widgets.line_wave_type);
	SET_GTK_SPIN_BUTTON		(vals->line_width,						widgets.line_width);
	SET_GTK_SPIN_BUTTON		(vals->line_width_rand,					widgets.line_width_rand);
	SET_GTK_SPIN_BUTTON		(vals->line_width_rand_seed,			widgets.line_width_rand_seed);
	SET_GTK_SPIN_BUTTON		(vals->line_in_margin,					widgets.line_in_margin);
	SET_GTK_SPIN_BUTTON		(vals->line_in_margin_rand,				widgets.line_in_margin_rand);
	SET_GTK_SPIN_BUTTON		(vals->line_in_margin_rand_seed,		widgets.line_in_margin_rand_seed);
	SET_GTK_SPIN_BUTTON		(vals->line_in_length,					widgets.line_in_length);
	SET_GTK_SPIN_BUTTON		(vals->line_in_length_rand,				widgets.line_in_length_rand);
	SET_GTK_SPIN_BUTTON		(vals->line_in_length_rand_seed,		widgets.line_in_length_rand_seed);
	SET_GTK_SPIN_BUTTON		(vals->line_out_margin,					widgets.line_out_margin);
	SET_GTK_SPIN_BUTTON		(vals->line_out_margin_rand,			widgets.line_out_margin_rand);
	SET_GTK_SPIN_BUTTON		(vals->line_out_margin_rand_seed,		widgets.line_out_margin_rand_seed);
	SET_GTK_SPIN_BUTTON		(vals->line_out_length,					widgets.line_out_length);
	SET_GTK_SPIN_BUTTON		(vals->line_out_length_rand,			widgets.line_out_length_rand);
	SET_GTK_SPIN_BUTTON		(vals->line_out_length_rand_seed,		widgets.line_out_length_rand_seed);

	{
		GtkWidget *temp = NULL;
		switch (vals->area_type) {
			case AREA_TYPE_RADIATION: temp = widgets.area_type_radiation; break;
			case AREA_TYPE_PARALLEL:  temp = widgets.area_type_parallel; break;
		}
		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (temp), TRUE);
		gtk_notebook_set_current_page (GTK_NOTEBOOK (widgets.area_type_notebook), vals->area_type);
	}
	SET_GIMP_INT_COMBO_BOX (vals->area_unit,           widgets.area_type_unit);

	SET_GTK_SPIN_BUTTON (vals->radiation_x,          widgets.radiation_x);
	SET_GTK_SPIN_BUTTON (vals->radiation_y,          widgets.radiation_y);
	SET_GTK_SPIN_BUTTON (vals->radiation_in_radius,  widgets.radiation_in_radius);
	SET_GTK_SPIN_BUTTON (vals->radiation_out_radius, widgets.radiation_out_radius);
	SET_GTK_SPIN_BUTTON (vals->radiation_oblateness, widgets.radiation_oblateness);
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widgets.radiation_in_freehand), vals->radiation_in_freehand_active);
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widgets.radiation_out_freehand), vals->radiation_out_freehand_active);

	SET_GTK_SPIN_BUTTON (vals->parallel_in_x,      widgets.parallel_in_x);
	SET_GTK_SPIN_BUTTON (vals->parallel_in_y,      widgets.parallel_in_y);
	SET_GTK_SPIN_BUTTON (vals->parallel_in_range,  widgets.parallel_in_range);
	SET_GTK_SPIN_BUTTON (vals->parallel_out_x,     widgets.parallel_out_x);
	SET_GTK_SPIN_BUTTON (vals->parallel_out_y,     widgets.parallel_out_y);
	SET_GTK_SPIN_BUTTON (vals->parallel_out_range, widgets.parallel_out_range);
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widgets.parallel_in_freehand), vals->parallel_in_freehand_active);
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widgets.parallel_out_freehand), vals->parallel_out_freehand_active);

	gtk_widget_queue_draw (widgets.preview);
	return;
#undef SET_GIMP_INT_COMBO_BOX
#undef SET_GTK_SPIN_BUTTON
}/*}}}*/

static void
keyfile_to_widgets (GKeyFile *kfile)
{/*{{{*/
	PluginVals *vals;
	vals = keyfile_to_pluginvals (kfile);
	pluginvals_to_widgets (vals);
	g_free (vals);
}/*}}}*/

MAIN()

/* vim: set fdm=marker ts=4 sw=4 noet :*/
