/*
 * GIMP Speedline Plug-in -- This is a plug-in for The GIMP 2
 * Copyright (C) 2013-2015 OHTSUKA, Yoshio <ohtsuka.yoshio@gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */
#include "speedline_line.h"

LineObj *
line_new (PluginVals *vals)
{
	gdouble resx, resy;
	LineObj *line = g_new0 (LineObj, 1);

	gimp_image_get_resolution (vals->image_ID, &resx, &resy);

	line->antialias = vals->line_antialias;

	line->color = vals->line_color;

	line->pitch           = gimp_units_to_pixels (vals->line_pitch,           vals->area_unit, resx);
	line->block_on_pitch  = gimp_units_to_pixels (vals->line_block_on_pitch,  vals->area_unit, resx);
	line->block_off_pitch = gimp_units_to_pixels (vals->line_block_off_pitch, vals->area_unit, resx);
	line->wave_npoint     = vals->line_wave_npoint;
	line->wave_depth      = gimp_units_to_pixels (vals->line_wave_depth,      vals->area_unit, resx);
	line->width           = gimp_units_to_pixels (vals->line_width,           vals->area_unit, resx);
	line->wave_type       = vals->line_wave_type;
	line->in_margin       = gimp_units_to_pixels (vals->line_in_margin,       vals->area_unit, resx);
	line->in_length       = gimp_units_to_pixels (vals->line_in_length,       vals->area_unit, resx);
	line->out_margin      = gimp_units_to_pixels (vals->line_out_margin,      vals->area_unit, resx);
	line->out_length      = gimp_units_to_pixels (vals->line_out_length,      vals->area_unit, resx);

	line->block_state     = 0;

	line->pitch_rand           = vals->line_pitch_rand;
	line->block_on_pitch_rand  = vals->line_block_on_pitch_rand;
	line->block_off_pitch_rand = vals->line_block_off_pitch_rand;
	line->wave_depth_rand      = vals->line_wave_depth_rand;
	line->width_rand           = vals->line_width_rand;
	line->in_margin_rand       = vals->line_in_margin_rand;
	line->in_length_rand       = vals->line_in_length_rand;
	line->out_margin_rand      = vals->line_out_margin_rand;
	line->out_length_rand      = vals->line_out_length_rand;

	line->randobj_pitch           = g_rand_new_with_seed (vals->line_pitch_rand_seed);
	line->randobj_block_on_pitch  = g_rand_new_with_seed (vals->line_block_on_pitch_rand_seed);
	line->randobj_block_off_pitch = g_rand_new_with_seed (vals->line_block_off_pitch_rand_seed);
	line->randobj_wave_depth      = g_rand_new_with_seed (vals->line_wave_depth_rand_seed);
	line->randobj_width           = g_rand_new_with_seed (vals->line_width_rand_seed);
	line->randobj_in_margin       = g_rand_new_with_seed (vals->line_in_margin_rand_seed);
	line->randobj_in_length       = g_rand_new_with_seed (vals->line_in_length_rand_seed);
	line->randobj_out_margin      = g_rand_new_with_seed (vals->line_out_margin_rand_seed);
	line->randobj_out_length      = g_rand_new_with_seed (vals->line_out_length_rand_seed);

	return line;
}

//draw a line at normalized coordinates; IN(0,0) -> OUT(0,length)
void
line_draw (LineObj	*line, 
		   cairo_t	*cr, 
		   double	length)
{
	double x0, y0, x1, y1, x2, y2, x3, y3;
	double w = line->width - (line->width * g_rand_double (line->randobj_width) * line->width_rand / 100.0);
	double in_margin  = line->in_margin  - (line->in_margin  * g_rand_double (line->randobj_in_margin)  * line->in_margin_rand  / 100.0);
	double in_length  = line->in_length  - (line->in_length  * g_rand_double (line->randobj_in_length)  * line->in_length_rand  / 100.0);
	double out_margin = line->out_margin - (line->out_margin * g_rand_double (line->randobj_out_margin) * line->out_margin_rand / 100.0);
	double out_length = line->out_length - (line->out_length * g_rand_double (line->randobj_out_length) * line->out_length_rand / 100.0);

	x0 = in_margin;
	y0 = 0;
	x1 = in_margin + in_length;
	y1 = w / 2;
	x2 = length - out_margin;
	y2 = 0;
	x3 = length - out_margin - out_length;
	y3 = w / 2;

	cairo_save (cr);
	{
		cairo_set_source_rgba (cr,
			line->color.r,
			line->color.g,
			line->color.b,
			line->color.a);

		switch (line->antialias) {
			case ANTIALIAS_TYPE_NONE:
				cairo_set_antialias (cr, CAIRO_ANTIALIAS_NONE);
				break;
			case ANTIALIAS_TYPE_FAST:
				cairo_set_antialias (cr, CAIRO_ANTIALIAS_FAST | CAIRO_ANTIALIAS_GRAY);
				break;
			case ANTIALIAS_TYPE_BEST:
				cairo_set_antialias (cr, CAIRO_ANTIALIAS_BEST | CAIRO_ANTIALIAS_SUBPIXEL);
				break;
		}

		if (x0 >= x2) {
			//draw nothing
		} else if (x1 > x3) {
			//get cross point (x4,y4) of line1(x0,y0)(x1,y1) and line2(x2,y2)(x3,y3)
			//line1 : y = a1 * x + b1;
			//line2 : y = a2 * x + b2;
			double x4, y4;
			if (x1 == x0) {
				double a2 = (y3 - y2) / (x3 - x2);
				double b2 = y2 - a2 * x2;
				x4 = x0;
				y4 = a2 * x4 + b2;
			} else if (x3 == x2) {
				double a1 = (y1 - y0) / (x1 - x0);
				double b1 = y1 - a1 * x1;
				x4 = x3;
				y4 = a1 * x4 + b1;
			} else {
				double a1 = (y1 - y0) / (x1 - x0);
				double b1 = y1 - a1 * x1;
				double a2 = (y3 - y2) / (x3 - x2);
				double b2 = y2 - a2 * x2;
				x4 = (b2 - b1) / (a1 - a2);
				y4 = a1 * x4 + b1;
			}
			cairo_move_to (cr, x0,  y0);
			cairo_line_to (cr, x4,  y4);
			cairo_line_to (cr, x2,  y2);
			cairo_line_to (cr, x4, -y4);
			cairo_close_path (cr);
			cairo_fill (cr);
		} else {
			cairo_move_to (cr, x0,  y0);
			cairo_line_to (cr, x1,  y1);
			cairo_line_to (cr, x3,  y3);
			cairo_line_to (cr, x2,  y2);
			cairo_line_to (cr, x3, -y3);
			cairo_line_to (cr, x1, -y1);
			cairo_close_path (cr);
			cairo_fill (cr);
		}
	}
	cairo_restore (cr);
}

gboolean
line_is_block_mode (LineObj *line)
{
	return line->block_on_pitch > 0 && line->block_off_pitch > 0;
}

gboolean
line_is_wave_mode (LineObj *line)
{
	return line->wave_npoint > 0 && line->wave_depth > 0;
}

void
line_block_state_set (LineObj *line,
					  gboolean val)
{
	line->block_state = val;
}

gboolean line_block_state_get (LineObj *line)
{
	return line->block_state;
}

gboolean
line_block_state_invert (LineObj *line)
{

	line->block_state ^= 1;
	return line->block_state;
}

gdouble
line_block_on_pitch_get (LineObj *line)
{
	gdouble pitch;
	pitch = line->block_on_pitch;
	return pitch - (pitch * g_rand_double (line->randobj_block_on_pitch) * line->block_on_pitch_rand / 100.0);
}

gdouble
line_block_off_pitch_get (LineObj *line)
{
	gdouble pitch;
	pitch = line->block_off_pitch;
	return pitch - (pitch * g_rand_double (line->randobj_block_off_pitch) * line->block_off_pitch_rand / 100.0);
}

gint
line_wave_npoint_get (LineObj *line)
{
	return line->wave_npoint;
}

gdouble
line_wave_depth_get (LineObj *line, gdouble n)
{
	gdouble depth;
	gdouble rate;
	
	rate = line_wave_depth_rate_get (line, n);
	depth = line->wave_depth * rate;
	return depth - (depth * g_rand_double (line->randobj_wave_depth) * line->wave_depth_rand / 100.0);
}

gdouble
line_wave_depth_rate_get (LineObj *line,
						  gdouble n)
{
	if (n > 1) {
		n = n - (int)n;
	}
	if (n < 0) {
		n = 1 + n - (int)n;
	}

	switch (line->wave_type) {
		case LINE_WAVE_TYPE_DEFAULT:
		case LINE_WAVE_TYPE_DEFAULT_INVERT:
			{
				double val;
				// y = 2x + 0
				if (n < 0.5) {
					val = 2 * n;
				} else if (n > 0.5) {
				// y = -2x + 2;
					val = -2 * n + 2;
				} else {
					val = 1;
				}
				if (line->wave_type == LINE_WAVE_TYPE_DEFAULT) {
					return val;
				} else {
					return 1 - val;
				}
			}

		case LINE_WAVE_TYPE_COS:
			return (1 + cos (G_PI * 2 * n)) / 2;

		case LINE_WAVE_TYPE_COS_INVERT:
			return (1 - cos (G_PI * 2 * n)) / 2;

		case LINE_WAVE_TYPE_ARC:
			return sqrt (1 - pow (2 * n -1, 2));

		case LINE_WAVE_TYPE_ARC_INVERT:
			return 1 - sqrt (1 - pow (2 * n -1, 2));

		case LINE_WAVE_TYPE_N:
			return n;

		case LINE_WAVE_TYPE_N_INVERT:
			return 1 - n;

		case LINE_WAVE_TYPE_BLOCK:
		case LINE_WAVE_TYPE_BLOCK_INVERT:
			if (n < 0.25 || n > 0.75) {
				return line->wave_type == LINE_WAVE_TYPE_BLOCK ? 1 : 0;
			} else {
				return line->wave_type == LINE_WAVE_TYPE_BLOCK ? 0 : 1;
			}

		default:
			return 1;
	}
	return 1;
}

void
line_destroy (LineObj *line)
{
	if (line->randobj_pitch)            g_free (line->randobj_pitch);
	if (line->randobj_block_on_pitch)   g_free (line->randobj_block_on_pitch);
	if (line->randobj_block_off_pitch)  g_free (line->randobj_block_off_pitch);
	if (line->randobj_wave_depth)       g_free (line->randobj_wave_depth);

	if (line->randobj_width)      g_free (line->randobj_width);
	if (line->randobj_in_margin)  g_free (line->randobj_in_margin);
	if (line->randobj_in_length)  g_free (line->randobj_in_length);
	if (line->randobj_out_margin) g_free (line->randobj_out_margin);
	if (line->randobj_out_length) g_free (line->randobj_out_length);

	line->randobj_pitch           = NULL;
	line->randobj_block_on_pitch  = NULL;
	line->randobj_block_off_pitch = NULL;
	line->randobj_wave_depth      = NULL;
	line->randobj_width           = NULL;
	line->randobj_in_margin       = NULL;
	line->randobj_in_length       = NULL;
	line->randobj_out_margin      = NULL;
	line->randobj_out_length      = NULL;

	if (line) g_free (line);
}

/* vim: set fdm=marker ts=4 sw=4 noet :*/
