/*
 * GIMP Speedline Plug-in -- This is a plug-in for The GIMP 2
 * Copyright (C) 2013-2015 OHTSUKA, Yoshio <ohtsuka.yoshio@gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */
#ifndef SPEEDLINE_LINE_H
#define SPEEDLINE_LINE_H
#include "speedline.h"

typedef struct _LineObj {
	GimpRGB			color;
	AntialiasType	antialias;
	gdouble			pitch;
	gdouble			block_on_pitch;
	gdouble			block_off_pitch;
	gint			wave_npoint;
	gdouble			wave_depth;
	LineWaveType	wave_type;
	gdouble			width;
	gdouble			in_margin;
	gdouble			out_margin;
	gdouble			in_length;
	gdouble			out_length;

	gboolean		block_state;

	gdouble			pitch_rand;
	gdouble			block_on_pitch_rand;
	gdouble			block_off_pitch_rand;
	gdouble			wave_depth_rand;
	gdouble			width_rand;
	gdouble			in_margin_rand;
	gdouble			out_margin_rand;
	gdouble			in_length_rand;
	gdouble			out_length_rand;

	GRand			*randobj_pitch;
	GRand			*randobj_block_on_pitch;
	GRand			*randobj_block_off_pitch;
	GRand			*randobj_wave_depth;
	GRand			*randobj_width;
	GRand			*randobj_in_margin;
	GRand			*randobj_out_margin;
	GRand			*randobj_in_length;
	GRand			*randobj_out_length;
} LineObj;

LineObj *line_new (PluginVals *vals);
void     line_destroy (LineObj *line);
void     line_draw (LineObj *line, cairo_t *cr, double length);
gboolean line_is_block_mode (LineObj *line);
gboolean line_is_wave_mode (LineObj *line);
void     line_block_state_set (LineObj *line, gboolean val);
gboolean line_block_state_get (LineObj *line);
gboolean line_block_state_invert (LineObj *line);
gdouble  line_block_on_pitch_get (LineObj *line);
gdouble  line_block_off_pitch_get (LineObj *line);
gint     line_wave_npoint_get (LineObj *line);
gdouble  line_wave_depth_get (LineObj *line, gdouble n);
gdouble  line_wave_depth_rate_get (LineObj *line, gdouble n);

#endif
/* vim: set fdm=marker ts=4 sw=4 noet :*/
